#include "ClassMember.h"
#include "Optimizer.h"
#include "ConceptInterpreter.h"
#include "SHManager.h"

//#define EMPIRIC_STACK_CHECK
POOLED_IMPLEMENTATION(ExecutableMember)
POOLED_IMPLEMENTATION(ClassMember)

ClassMember::ClassMember(void *DefinedIn, char *name, /*AnsiString Filename,*/ char is_data_only, char _not_binary, char is_unserialize) {
    // do not own the name !
    NAME   = name;
    ACCESS = ACCESS_PUBLIC;

    IS_FUNCTION           = 0;
    IS_STATIC             = 0;
    IS_OPERATOR           = 0;
    MEMBER_SET            = 0;
    MEMBER_GET            = 0;
    OPTIMIZER             = 0;
    PARAMETERS_COUNT      = 0;
    MUST_PARAMETERS_COUNT = 0;

    Defined_In       = DefinedIn;
    _DEBUG_STARTLINE = 0;

    VD = 0;
    if (is_data_only) {
        if (!is_unserialize) {
            VD = new SmallVariableDESCRIPTOR;
            VD->TYPE   = VARIABLE_NUMBER;
            VD->value  = "0";
            VD->nValue = 0;
        }
        VariableDescriptors = 0;
        PIF_DATA            = 0;
    } else {
        if (_not_binary) {
            PIF_DATA            = new DoubleList();
            VariableDescriptors = new DoubleList();
        } else {
            PIF_DATA = 0;
            // april 6, 2010
            VariableDescriptors = 0;
        }
    }
}

void ClassMember::EnsureVD() {
    if (VD) {
        return;
    }
    VD = new SmallVariableDESCRIPTOR;
    VD->TYPE   = VARIABLE_NUMBER;
    VD->value  = "0";
    VD->nValue = 0;
}

int ClassMember::Serialize(FILE *out, bool is_lib) {
    //NAME.Serialize(out,SERIALIZE_16BIT_LENGTH);
    concept_fwrite(&ACCESS, sizeof(ACCESS), 1, out);
    char is_funtion_or_operator = IS_FUNCTION;
    if (IS_OPERATOR) {
        is_funtion_or_operator = (signed char)-1;
    }

    concept_fwrite(&is_funtion_or_operator, sizeof(is_funtion_or_operator), 1, out);
    if (IS_FUNCTION == 2) {
        concept_fwrite(&MEMBER_GET, sizeof(MEMBER_GET), 1, out);
    } else
    if (IS_FUNCTION == 3) {
        concept_fwrite(&MEMBER_SET, sizeof(MEMBER_SET), 1, out);
        concept_fwrite(&MEMBER_GET, sizeof(MEMBER_GET), 1, out);
    } else
    if (IS_FUNCTION == 1) {
        concept_fwrite(&PARAMETERS_COUNT, sizeof(PARAMETERS_COUNT), 1, out);
        concept_fwrite(&MUST_PARAMETERS_COUNT, sizeof(PARAMETERS_COUNT), 1, out);
        concept_fwrite(&IS_STATIC, sizeof(IS_STATIC), 1, out);
    }

    SmallVariableDESCRIPTOR DUMMY;
    DUMMY.nValue = 0;
    DUMMY.TYPE   = VARIABLE_NUMBER;

    SmallVariableDESCRIPTOR *refVD = &DUMMY;
    if (VD) {
        refVD = VD;
    }

    SERIALIZE_SMALL_VAR_DESCRIPTOR(refVD, out);
    if (IS_FUNCTION == 1) {
        ((Optimizer *)OPTIMIZER)->Serialize(out, (INTEGER)is_lib);
    }
    return(0);
}

int ClassMember::ComputeSharedSize(concept_FILE *in, signed char& IS_FUNCTION) {
    int  size = 0; //sizeof(ClassMember);
    signed char ACCESS;
    signed char IS_OPERATOR;

    FREAD_FAIL(&ACCESS, sizeof(ACCESS), 1, in);
    FREAD_FAIL(&IS_FUNCTION, sizeof(IS_FUNCTION), 1, in);
    if (IS_FUNCTION < (signed char)0) {
        IS_OPERATOR = 1;
        IS_FUNCTION = 1;
    }
    if (IS_FUNCTION == 2) {
        // MEMBER_GET;
        SKIP(sizeof(INTEGER), in);
    } else
    if (IS_FUNCTION == 3) {
        // MEMBER_GET
        SKIP(sizeof(INTEGER), in);
        // MEMBER_SET
        SKIP(sizeof(INTEGER), in);
    } else
    if (IS_FUNCTION == 1) {
        //PARAMETERS_COUNT
        SKIP(sizeof(unsigned short), in)
        //MUST_PARAMETERS_COUNT
        SKIP(sizeof(unsigned short), in);
        // daca este functie, atunci citesc si "IS_STATIC"
        SKIP(1, in);
        size += sizeof(AnsiList);                    //VariableDescriptors
    }
    VariableDESCRIPTOR VD;
    UNSERIALIZE_VAR_DESCRIPTOR_SIZE((&VD), in);
    int vd_size = VD.value.Length();
    if (vd_size) {
        size += vd_size + 1;
    }
#ifdef ARM_ADJUST_SIZE
    return ARM_ADJUST_SIZE(size);
#else
    return(size);
#endif
}

int ClassMember::Unserialize(concept_FILE *in, AnsiList *ClassList /*, AnsiList *ModuleList*/, bool is_lib, int *Relocation, int *ClassNames) {
    concept_fread(&ACCESS, sizeof(ACCESS), 1, in);
    concept_fread(&IS_FUNCTION, sizeof(IS_FUNCTION), 1, in);
    if (IS_FUNCTION < (signed char)0) {
        IS_OPERATOR = 1;
        IS_FUNCTION = 1;
    }

    if (IS_FUNCTION == 2) {
        concept_fread(&MEMBER_GET, sizeof(MEMBER_GET), 1, in);
        if (is_lib) {
            MEMBER_GET = Relocation [MEMBER_GET - 1] + 1;
        }
    } else
    if (IS_FUNCTION == 3) {
        concept_fread(&MEMBER_SET, sizeof(MEMBER_SET), 1, in);
        concept_fread(&MEMBER_GET, sizeof(MEMBER_GET), 1, in);
        if (is_lib) {
            if (MEMBER_SET) {
                MEMBER_SET = Relocation [MEMBER_SET - 1] + 1;
            }
            if (MEMBER_GET) {
                MEMBER_GET = Relocation [MEMBER_GET - 1] + 1;
            }
        }
    } else
    if (IS_FUNCTION == 1) {
        concept_fread(&PARAMETERS_COUNT, sizeof(PARAMETERS_COUNT), 1, in);
        concept_fread(&MUST_PARAMETERS_COUNT, sizeof(PARAMETERS_COUNT), 1, in);
        // daca este functie, atunci citesc si "IS_STATIC"
        concept_fread(&IS_STATIC, sizeof(IS_STATIC), 1, in);
    }

    SmallVariableDESCRIPTOR DUMMY;
    DUMMY.nValue = 0;
    DUMMY.TYPE   = VARIABLE_NUMBER;

    SmallVariableDESCRIPTOR *refVD = &DUMMY;
    if (!IS_FUNCTION) {
        this->EnsureVD();
        refVD = VD;
    } else {
        VD = 0;
    }

    if (!IS_FUNCTION) {
        refVD->value = (char *)"";
    }

    UNSERIALIZE_SMALL_VAR_DESCRIPTOR(refVD, in, false);
    if ((refVD->TYPE==TYPE_NUMBER) && (!refVD->nValue) && (VD)) {
        delete VD;
        VD=0;
    }
    return(0);
}

ClassMember::~ClassMember(void) {
    if (OPTIMIZER) {
        if (MEMBER_GET != -1) {
            delete (Optimizer *)OPTIMIZER;
        }
    }

    if (VariableDescriptors) {
        delete VariableDescriptors;
    }

    if (PIF_DATA) {
        delete PIF_DATA;
    }

    if (VD) {
        delete VD;
    }
}

#ifdef EMPIRIC_STACK_CHECK
static int STACK_HIT = 0;
 #define MAX_RECURSIVE_CALL    2100
#endif

VariableDATA *ClassMember::Execute(intptr_t CONCEPT_CLASS_ID, VariableDATA *Owner, ParamList *FORMAL_PARAM, VariableDATA **SenderCTX, VariableDATA *& THROW_DATA, char is_main) {
#ifdef EMPIRIC_STACK_CHECK
    if (++STACK_HIT > MAX_RECURSIVE_CALL) {
        ((Optimizer *)OPTIMIZER)->PIFOwner->Errors.Add(new AnsiException(ERR840, _DEBUG_STARTLINE, 840, NAME, _DEBUG_FILENAME), DATA_EXCEPTION);
        return(0);
    }
#endif

    ConceptInterpreter *INTERPRETER=(ConceptInterpreter *)((Optimizer *)this->OPTIMIZER)->INTERPRETER;
    if (!INTERPRETER) {
        INTERPRETER = new ConceptInterpreter((Optimizer *)OPTIMIZER, ((Optimizer *)this->OPTIMIZER)->PIFOwner->out, &((Optimizer *)this->OPTIMIZER)->PIFOwner->ClassList, ((ClassCode *)Defined_In)->CLSID, this, ((Optimizer *)this->OPTIMIZER)->PIFOwner->DEBUGGER_TRAP, ((Optimizer *)this->OPTIMIZER)->PIFOwner->DEBUGGER_RESERVED, ((Optimizer *)this->OPTIMIZER)->PIFOwner->SERVER_PUBLIC_KEY, ((Optimizer *)this->OPTIMIZER)->PIFOwner->SERVER_PRIVATE_KEY, ((Optimizer *)this->OPTIMIZER)->PIFOwner->CLIENT_PUBLIC_KEY);
        ((Optimizer *)this->OPTIMIZER)->INTERPRETER = INTERPRETER;
    }

#ifdef USE_JIT
    if (INTERPRETER->JITCode) {
        if (INTERPRETER->JITCode==-1) {
            if (INTERPRETER->CanJIT())
                INTERPRETER->JITCode=(SYS_INT)INTERPRETER->JIT();
            else
                INTERPRETER->JITCode=0;
        }

        void *code=(void *)INTERPRETER->JITCode;
        if (code) {
            THROW_DATA=NULL;

            VariableDATA *RESULT=(VariableDATA *)malloc(sizeof(VariableDATA));
            RESULT->CALL_SET=0;
            RESULT->CLASS_DATA=0;
            RESULT->IS_PROPERTY_RESULT=0;
            RESULT->LINKS=0;
            RESULT->TYPE=VARIABLE_NUMBER;
            RESULT->NUMBER_DATA=INTERPRETER->JITExecute(code, FORMAL_PARAM, SenderCTX);

            //INTERPRETER->JITDone(code);
            return RESULT;
        }
    }
#endif
    ((ConceptInterpreter *)INTERPRETER)->static_call_main=is_main;
    VariableDATA **CONTEXT = ((ConceptInterpreter *)INTERPRETER)->CreateEnviroment(Owner, FORMAL_PARAM, SenderCTX);
    VariableDATA *RESULT   = ((ConceptInterpreter *)INTERPRETER)->Interpret(CONTEXT, CONCEPT_CLASS_ID, THROW_DATA);
    if (CONTEXT) {
        ((ConceptInterpreter *)INTERPRETER)->DestroyEnviroment(CONTEXT);
    }
#ifdef EMPIRIC_STACK_CHECK
    --STACK_HIT;
#endif
    if (RESULT) {
        RESULT->LINKS--;
    }
    return(RESULT);
}

GreenThreadCycle *ClassMember::CreateThread(intptr_t CONCEPT_CLASS_ID, VariableDATA *Owner) {
    // cannot query anthing else except functions
    if (this->IS_FUNCTION!=1)
        return 0;
    ConceptInterpreter *INTERPRETER=(ConceptInterpreter *)((Optimizer *)this->OPTIMIZER)->INTERPRETER;
    if (!INTERPRETER) {
        INTERPRETER = new ConceptInterpreter((Optimizer *)OPTIMIZER, ((Optimizer *)this->OPTIMIZER)->PIFOwner->out, &((Optimizer *)this->OPTIMIZER)->PIFOwner->ClassList, ((ClassCode *)Defined_In)->CLSID, this, ((Optimizer *)this->OPTIMIZER)->PIFOwner->DEBUGGER_TRAP, ((Optimizer *)this->OPTIMIZER)->PIFOwner->DEBUGGER_RESERVED, ((Optimizer *)this->OPTIMIZER)->PIFOwner->SERVER_PUBLIC_KEY, ((Optimizer *)this->OPTIMIZER)->PIFOwner->SERVER_PRIVATE_KEY, ((Optimizer *)this->OPTIMIZER)->PIFOwner->CLIENT_PUBLIC_KEY);
        ((Optimizer *)this->OPTIMIZER)->INTERPRETER = INTERPRETER;
    }

    ParamList FORMAL_PARAM;
    FORMAL_PARAM.COUNT       = 0;
    FORMAL_PARAM.PARAM_INDEX = 0;

    VariableDATA **CONTEXT = ((ConceptInterpreter *)INTERPRETER)->CreateEnviroment(Owner, &FORMAL_PARAM, 0);

    GreenThreadCycle *gtc=(GreenThreadCycle *)malloc(sizeof(GreenThreadCycle));
    if (gtc) {
        gtc->CATCH_INSTRUCTION_POINTER = 0;
        gtc->CATCH_VARIABLE = 0;
        gtc->INSTRUCTION_COUNT = ((Optimizer *)this->OPTIMIZER)->codeCount;
        gtc->INSTRUCTION_POINTER = 0;
        gtc->INTERPRETER = INTERPRETER;
        gtc->LOCAL_CONTEXT = CONTEXT;
        gtc->NEXT = 0;
        gtc->PREVIOUS_TRY = 0;
        gtc->THROW_DATA = 0;
        gtc->LAST_THREAD = 0;
        gtc->CURRENT_THREAD = 0;
        gtc->OWNER = Owner;
    }
    return gtc;
}

void ClassMember::DoneThread(GreenThreadCycle *gtc) {
    if (gtc) {
        if (gtc->THROW_DATA) {
            FREE_VARIABLE(gtc->THROW_DATA);
        }

        if (gtc->LOCAL_CONTEXT)
            ((ConceptInterpreter *)gtc->INTERPRETER)->DestroyEnviroment(gtc->LOCAL_CONTEXT);

        if (gtc->OWNER) {
            FREE_VARIABLE(gtc->OWNER);
        }

        free(gtc);
    }
}