#include "GarbageCollector.h"
#include "CompiledClass.h"
#include "ClassCode.h"
#include "Array.h"

POOLED_IMPLEMENTATION(_GarbageElement)
POOLED_IMPLEMENTATION(GarbageCollector)

#ifdef SIMPLE_MULTI_THREADING
HHSEM GarbageCollector::gc_lock;
#endif

GarbageCollector::GarbageCollector() {
    BASE = 0;
}

void GarbageCollector::Call_All_Destructors() {
#ifdef USE_MAP_GC
    if (BASE) {
        std::set<void *>::iterator end=BASE->end();
        for (std::set<void *>::iterator i = BASE->begin(); i!=end; ++i) {
            CompiledClass *ptr = (CompiledClass *)(*i);
            if (ptr->HasDestructor())
                ptr->Destroy();
        }
    }
#else
    GarbageElement *NODE  = BASE;
    GarbageElement *NODE2 = 0;

    while (NODE) {
        NODE2 = NODE->NEXT;
        CompiledClass *ptr = (CompiledClass *)NODE->DATA;

        if (ptr->HasDestructor()) {
            ptr->Destroy();
        }

        NODE = NODE2;
    }
#endif
}

void GarbageCollector::EndOfExecution_SayBye_Objects() {
#ifdef USE_MAP_GC
    if (BASE) {
        std::set<void *> *BASE2=BASE;
        BASE=0;

        std::set<void *>::iterator end=BASE2->end();
        for (std::set<void *>::iterator i = BASE2->begin(); i!=end; ++i) {
            CompiledClass *ptr = (CompiledClass *)(*i);
            ptr->LINKS++;

            ptr->UnlinkObjects();
            delete ptr;
        }
        BASE2->clear();
        delete BASE2;
    }
#else
    GarbageElement *NODE  = BASE;
    GarbageElement *NODE2 = 0;

    // to prevent recursive re-execution
    BASE = 0;
    while (NODE) {
        NODE2 = NODE->NEXT;
        CompiledClass *ptr = (CompiledClass *)NODE->DATA;
        free(NODE);
        NODE = NODE2;

        ptr->LINKS++;

        ptr->UnlinkObjects();
        delete ptr;
    }
#endif
}

void GarbageCollector::EndOfExecution_SayBye_Arrays() {
#ifdef USE_MAP_GC
    if (BASE) {
        std::set<void *> *BASE2=BASE;
        BASE=0;

        std::set<void *>::iterator end=BASE2->end();
        for (std::set<void *>::iterator i = BASE2->begin(); i!=end; ++i) {
            Array *ptr = (Array *)(*i);
            ptr->LINKS++;

            ptr->UnlinkObjects();
            delete ptr;
        }
        BASE2->clear();
        delete BASE2;
    }
#else
    GarbageElement *NODE  = BASE;
    GarbageElement *NODE2 = 0;

    BASE = 0;
    while (NODE) {
        NODE2 = NODE->NEXT;
        Array *ptr = (Array *)NODE->DATA;
        free(NODE);
        NODE = NODE2;

        ptr->LINKS++;

        ptr->UnlinkObjects();
        delete ptr;
    }
#endif
}

/*void GarbageCollector::Reference(void *data) {
 *  GarbageElement *NODE=(GarbageElement *)malloc(sizeof(GarbageElement));
 *  NODE->NEXT=BASE;
 *  NODE->DATA=data;
 *
 *  if (!BASE) {
 *      BASE=NODE;
 *  }
 * }*/

/*
 * void *GarbageCollector::Unreference(void *data) {
 *  GarbageElement *NODE=BASE;
 *  GarbageElement *NODE_PREC=0;
 *  while (NODE) {
 *      if (NODE->DATA==data) {
 *          if (NODE_PREC)
 *              NODE_PREC->NEXT=NODE->NEXT;
 *          else
 *              BASE=NODE->NEXT;
 *
 *          free(NODE);
 *          return data;
 *      }
 *      NODE_PREC=NODE;
 *      NODE=NODE->NEXT;
 *  }
 *  return 0;
 * }
 */
#ifdef SIMPLE_MULTI_THREADING
bool GarbageCollector::Lock() {
    if (ClassCode::PIF->ThreadsCount>0) {
        if (ClassCode::PIF->MasterLock)
            return false;
        semp(gc_lock);
        return true;
    }
    return false;
}

void GarbageCollector::Unlock() {
    semv(gc_lock);
}
#endif

GarbageCollector::~GarbageCollector() {
#ifdef USE_MAP_GC
    if (BASE) {
        BASE->clear();
        delete BASE;
        BASE=0;
    }
#else
    GarbageElement *NODE  = BASE;
    GarbageElement *NODE2 = 0;

    while (NODE) {
        NODE2 = NODE->NEXT;
        free(NODE);
        NODE = NODE2;
    }
#endif
}
