#ifndef __PIFALIZATOR_H
#define __PIFALIZATOR_H

#include "AnsiString.h"
#include "AnsiList.h"
#include "Codes.h"
#include "AnsiParser.h"
#include "ClassCode.h"
#include "Debugger.h"
#include "ConceptPools.h"
#include "TinyString.h"
#include "StaticList.h"

#include "semhh.h"

#ifdef SIMPLE_MULTI_THREADING

    #define CC_WRITE_LOCK(PIF)             bool IsWriteLocked=false; if (PIF->ThreadsCount>0) { PIF->MasterLock=1; semp(PIF->WriteLock); IsWriteLocked=true; }
    #define CC_WRITE_LOCK2(PIF)            IsWriteLocked=false; if (PIF->ThreadsCount>0) { PIF->MasterLock=1; semp(PIF->WriteLock); IsWriteLocked=true; }
    #define CC_WRITE_UNLOCK(PIF)           if (IsWriteLocked) { semv(PIF->WriteLock); PIF->MasterLock=0; IsWriteLocked=false; }

    #define WRITE_LOCK                     if ((!IsWriteLocked) && (PIF->ThreadsCount>0)) { semp(PIF->WriteLock); PIF->MasterLock=1; IsWriteLocked=1; }
    #define WRITE_UNLOCK                   if (IsWriteLocked) { IsWriteLocked=0; PIF->MasterLock=0; semv(PIF->WriteLock); }
    #define NEW_THREAD                     semp(PIF->InternalLock); PIF->ThreadsCount++; semv(PIF->InternalLock);
    #define DONE_THREAD                    semp(PIF->InternalLock); PIF->ThreadsCount--; semv(PIF->InternalLock);
    #define INTERNAL_LOCK(PIF)             semp(PIF->InternalLock);
    #define INTERNAL_UNLOCK(PIF)           semv(PIF->InternalLock);

    #define SMART_LOCK(VARIABLE)                                                                                                                        \
        if (VARIABLE->LINKS>1) {                                                                                                                        \
            WRITE_LOCK                                                                                                                                  \
        } else                                                                                                                                          \
        if (VARIABLE->CLASS_DATA) {                                                                                                                     \
            if (((VARIABLE->TYPE == VARIABLE_CLASS) || (VARIABLE->TYPE == VARIABLE_DELEGATE)) && (((CompiledClass *)VARIABLE->CLASS_DATA)->LINKS>1)) {  \
                WRITE_LOCK                                                                                                                              \
            } else                                                                                                                                      \
            if ((VARIABLE->TYPE == VARIABLE_ARRAY) && (((Array *)VARIABLE->CLASS_DATA)->LINKS>1)) {                                                     \
                WRITE_LOCK                                                                                                                              \
            }                                                                                                                                           \
        }   
#else
    #define CC_WRITE_LOCK(PIF)
    #define CC_WRITE_UNLOCK(PIF)
    #define CC_WRITE_LOCK2(PIF)

    #define WRITE_LOCK
    #define WRITE_UNLOCK
    #define NEW_THREAD
    #define DONE_THREAD
    #define INTERNAL_LOCK(PIF)
    #define INTERNAL_UNLOCK(PIF)

    #define SMART_LOCK(VARIABLE)
#endif

//#define CACHED_CLASSES
#define CACHED_VARIABLES

#ifdef CACHED_CLASSES
#include <map>
#endif

#ifdef CACHED_VARIABLES
#include <map>
#endif

#define PRAGMA_WARNINGS       (char *)"warnings"
#define PRAGMA_EXCEPTIONS     (char *)"lib_exceptions"
#define PRAGMA_IMPLICIT       (char *)"implicit"
#define PRAGMA_ENTRY_POINT    (char *)"entry_point"

#define PRAGMA_ON             (char *)"on"
#define PRAGMA_OFF            (char *)"off"


#define MAIN_ENTRY_POINT      "Main"
//#define DEFAULT_ENTRY_POINT           "Main"

//#define INCLUDE_DIR                 "Include/"
//#define IMPORT_DIR                  "Library/"
#define DEFAULT_USE_WARNINGS      1
#define DEFAULT_USE_EXCEPTIONS    0
#define DEFAULT_USE_IMPLICIT      0

#define DEFAULT_BIN_EXTENSION     ".accel"
#define DEFAULT_PACK_EXTENSION    ".package"

#define STATIC_LINKS_INCREMENT    0x20
class PIFAlizator {
    friend class Optimizer;
    friend class ClassCode;
    friend class ClassMember;
    friend class ConceptInterpreter;
    friend class CompiledClass;
    friend class GarbageCollector;
    friend INTEGER Invoke(INTEGER INVOKE_TYPE, ...);

    StaticList GeneralMembers;
    int        pipe_read;
    int        pipe_write;
    int        apid;
    int        parent_apid;

    SimpleStream *out;

    INTEGER DebugOn;

    AnsiString FileName;
    AnsiString *InputStream;

    AnsiList ClassList;

    AnsiList IncludedList;
    AnsiList ModuleList;
    AnsiList ModuleNamesList;
    AnsiList ConstantList;
    AnsiList UndefinedMembers;
    AnsiList UndefinedClasses;
    AnsiList DeletedMembers;

    SYS_INT      *StaticLinks;
    unsigned int StaticLinksCount;
    unsigned int StaticLinksSize;

    unsigned int LinkStatic(char *funname);
    INTEGER AddUndefinedMember(AnsiString& member, TinyString& _CLASS, char *_MEMBER, intptr_t line);

    //INTEGER AddGeneralMember(AnsiString name);
    //INTEGER DefineMember(AnsiString name);
    INTEGER CheckUndefinedMembers();

    INTEGER AddUndefinedClass(AnsiString& member, TinyString& _CLASS, char *_MEMBER, intptr_t line);
    INTEGER CheckUndefinedClasses();

#ifdef CACHED_VARIABLES
    INTEGER VariableIsDescribed(AnsiString& S, DoubleList *VDList, std::map<unsigned int, unsigned int> *CachedVariables=0, char is_hased=0);
#else
    INTEGER VariableIsDescribed(AnsiString& S, DoubleList *VDList);
#endif
    INTEGER ConstantIsDescribed(AnsiString& S, AnsiList *VDList);

    AnsiString GetSpecial(AnsiParser *P, ClassCode *CC, ClassMember *CM, AnsiString special);
    INTEGER BuildArray(ClassCode *CC, ClassMember *CM, AnsiParser *P, INTEGER on_line);
    INTEGER BuildFunction(ClassCode *CC, AnsiParser *P, INTEGER on_line = 0, INTEGER ACCESS = ACCESS_PUBLIC, INTEGER OPERATOR = 0, char STATIC = 0, char *prec_parse = 0, char is_inline = 0);
    INTEGER BuildProperty(ClassCode *CC, AnsiParser *P, INTEGER on_line = 0, INTEGER ACCESS = ACCESS_PUBLIC);
    INTEGER BuildEvent(ClassCode *CC, AnsiParser *P, INTEGER on_line = 0, INTEGER ACCESS = ACCESS_PUBLIC);
    INTEGER BuildOverride(ClassCode *CC, AnsiParser *P, INTEGER on_line);
    INTEGER IncludePackage(AnsiString filename);
    INTEGER IncludeFile(AnsiString MODULE_NAME, INTEGER on_line);

    AnsiString NormalizePath(AnsiString *MODULE_NAME);

    INTEGER BuildClass(AnsiParser *P, INTEGER on_line = 0);

    INTEGER BuildPragma(AnsiParser *P);

    INTEGER Execute(AnsiString *Stream, INTEGER on_line = 0, char _USE_WARN = DEFAULT_USE_WARNINGS, char _USE_EXC = DEFAULT_USE_EXCEPTIONS, char _USE_IMPLICIT = DEFAULT_USE_IMPLICIT);
    SYS_INT ClassExists(char *name, char by_addr = 0, int *index = 0);
    INTEGER ListContains(AnsiString& S, AnsiList *VDList, char is_tiny = false);
    INTEGER BuildVariable(ClassCode *CC, AnsiParser *P, INTEGER on_line, INTEGER ACCESS);
    INTEGER ExtendClass(ClassCode *CC, AnsiParser *P, INTEGER on_line, INTEGER OWNER_CLSID);

    INTEGER ENTRY_CLASS;

    DEBUGGER_CALLBACK DEBUGGER_TRAP;
    void              *DEBUGGER_RESERVED;
    char              *SERVER_PUBLIC_KEY;
    char              *SERVER_PRIVATE_KEY;
    char              *CLIENT_PUBLIC_KEY;

    AnsiString INCLUDE_DIR;
    AnsiString IMPORT_DIR;
    AnsiString TEMP_INC_DIR;

    char USE_WARN;
    char USE_EXC;
    char USE_IMPLICIT;
    char *PROFILE_DRIVEN;
    unsigned int PROFILE_DRIVEN_ID;
    int  INCLUDE_LEVEL;

    //AnsiString MAIN_ENTRY_POINT;

    INTEGER Warning(char *WRN, int line, int wrn_code, char *extra, char *filename = 0);
    void DefineConstant(char *name, char *value, int is_string = 1);

    char *CheckMember(char *member_name);

    //char basic_members_count;
    char basic_constants_count;

#ifdef CACHED_LIST
    std::map<unsigned int, unsigned int> CachedClasses;
#endif

#ifdef SIMPLE_MULTI_THREADING
    HHSEM WriteLock;
    char MasterLock;

    HHSEM InternalLock;
    uintptr_t ThreadsCount;
#endif
    HHSEM DelegateLock;

public:
    static int         refSOCKET;
    POOLED(PIFAlizator)
    void *CachedhDLL;
    static int         argc;
    static char        **argv;
    static int         last_result;
    static int         direct_pipe;
    static CHECK_POINT CheckPoint;
    static AnsiString GetPath(AnsiString *S);

    void AcknoledgeRunTimeError(AnsiException *Exc);

    DoubleList PIF;
    DoubleList VariableDescriptors;
    AnsiList Errors;
    AnsiList Warnings;

    AnsiString DEBUG_CLASS_CONFIGURATION();
    AnsiString DEBUG_INFO();
    AnsiString PRINT_ERRORS(int html = 0);
    AnsiString PRINT_WARNINGS(int html = 0);
    AnsiString SerializeWarningsErrors(int ser_warnings = 0);

    INTEGER ErrorCount();
    INTEGER WarningCount();

    void Adjust(INTEGER debug, DEBUGGER_CALLBACK DC, void *DEBUGGER_RESERVED, char *SPubK, char *SPrivK, char *CPubK);
    PIFAlizator(AnsiString INC_DIR, AnsiString LIB_DIR, AnsiString *S, SimpleStream *Out, AnsiString _FileName = "", INTEGER debug = 0, DEBUGGER_CALLBACK DC = 0, void *DEBUGGER_RESERVED = 0, char *SPubK = 0, char *SPrivK = 0, char *CPubK = 0);
    void SetPipe(int pipein, int pipeout, int apid, int papid);
    INTEGER Execute();
    void Optimize(char use_compiled_code = 0);
    void OptimizeMemoryUsage();
    bool CheckRunable();
    void *GetStartingPoint();
    int Serialize(char *filename, bool is_lib = false);
    int Unserialize(char *filename, bool is_lib = false);
    static int ComputeSharedSize(char *filename);
    void Hibernate();

    void Clear();
    ~PIFAlizator(void);
};

#endif //__PIFALIZATOR_H
