//------------ standard header -----------------------------------//
#include "stdlibrary.h"
//------------ end of standard header ----------------------------//
#include "library.h"
#include "AnsiString.h"
#include "interface.h"
#ifdef _WIN32
    #ifdef _WIN32_WINNT
        #undef _WIN32_WINNT
    #endif
    #define _WIN32_WINNT    0x501

    #define _WINSOCKAPI_
    #include <io.h>
    #include <winsock2.h>
    #include <ws2tcpip.h>
    #include <map>

    std::map<SYS_INT,int> UnixSocketEmulation;
    const char *inet_ntop(int af, const void *src, char *dst, socklen_t cnt) {
        if (af == AF_INET) {
            struct sockaddr_in in;
            memset(&in, 0, sizeof(in));
            in.sin_family = AF_INET;
            memcpy(&in.sin_addr, src, sizeof(struct in_addr));
            getnameinfo((struct sockaddr *)&in, sizeof(struct sockaddr_in), dst, cnt, NULL, 0, NI_NUMERICHOST);
            return(dst);
        } else if (af == AF_INET6)   {
            struct sockaddr_in6 in;
            memset(&in, 0, sizeof(in));
            in.sin6_family = AF_INET6;
            memcpy(&in.sin6_addr, src, sizeof(struct in_addr6));
            getnameinfo((struct sockaddr *)&in, sizeof(struct sockaddr_in6), dst, cnt, NULL, 0, NI_NUMERICHOST);
            return(dst);
        }
        return(NULL);
    }
#else
    #include <unistd.h>
    #include <sys/types.h>
    #include <sys/socket.h>
    #include <sys/un.h>
    #include <netinet/in.h>
    #include <netdb.h>
    #include <string.h>
    #include <arpa/inet.h>
    #include <netinet/tcp.h>
    #include <sys/select.h>
    #include <signal.h>

    #define INVALID_SOCKET	-1

    int eof(int stream) {
        struct timeval timeout;
        timeout.tv_sec = 0;
        timeout.tv_usec = 10;
        fd_set s;
        FD_ZERO(&s);
        FD_SET(stream,&s);

        int val=select(stream+1, &s, 0, 0, &timeout);

        if (val) {
           if (FD_ISSET(stream, &s))
               return 0;
        }
        return 1;
    }

    #ifndef UNIX_PATH_MAX
        #define UNIX_PATH_MAX 108
    #endif
#endif
#include "websocket.h"
#include "ping.h"
//=====================================================================================//
CONCEPT_DLL_API ON_CREATE_CONTEXT MANAGEMENT_PARAMETERS {
#ifdef _WIN32
    WSADATA wsaData;
    WSAStartup( MAKEWORD( 2, 2 ), &wsaData);
#else
    #ifndef SO_NOSIGPIPE
        signal(SIGPIPE, SIG_IGN);
    #endif
#endif
    Invoke(INVOKE_DEFINE_CONSTANT, HANDLER, "PROTOCOL_TCP","0");
    Invoke(INVOKE_DEFINE_CONSTANT, HANDLER, "PROTOCOL_UDP","1");
    Invoke(INVOKE_DEFINE_CONSTANT, HANDLER, "PROTOCOL_UNIX","2");
    Invoke(INVOKE_DEFINE_CONSTANT, HANDLER, "PROTOCOL_UNIXDGRAM","3");

    DEFINE_ECONSTANT(WS_ERROR_FRAME)
    DEFINE_ECONSTANT(WS_INCOMPLETE_FRAME)
    DEFINE_ECONSTANT(WS_TEXT_FRAME)
    DEFINE_ECONSTANT(WS_BINARY_FRAME)
    DEFINE_ECONSTANT(WS_OPENING_FRAME)
    DEFINE_ECONSTANT(WS_CLOSING_FRAME)
    DEFINE_ECONSTANT(WS_UNSUPPORTED_VERSION)        

    DEFINE_ECONSTANT(SOL_SOCKET)
    DEFINE_ECONSTANT(IPPROTO_IP)
    DEFINE_ECONSTANT(IPPROTO_TCP)
    DEFINE_ECONSTANT(IPPROTO_UDP)

    DEFINE_ECONSTANT(SO_DEBUG)
    DEFINE_ECONSTANT(SO_ACCEPTCONN)
    DEFINE_ECONSTANT(SO_REUSEADDR)
    DEFINE_ECONSTANT(SO_KEEPALIVE)
    DEFINE_ECONSTANT(SO_DONTROUTE)
    DEFINE_ECONSTANT(SO_BROADCAST)
    //DEFINE_ECONSTANT(SO_USELOOPBACK)
    DEFINE_ECONSTANT(SO_LINGER)
    DEFINE_ECONSTANT(SO_OOBINLINE)

    DEFINE_ECONSTANT(SO_SNDBUF)
    DEFINE_ECONSTANT(SO_RCVBUF)
    DEFINE_ECONSTANT(SO_SNDLOWAT)
    DEFINE_ECONSTANT(SO_RCVLOWAT)
    DEFINE_ECONSTANT(SO_SNDTIMEO)
    DEFINE_ECONSTANT(SO_RCVTIMEO)
    DEFINE_ECONSTANT(SO_ERROR)
    DEFINE_ECONSTANT(SO_TYPE)

    DEFINE_ECONSTANT(TCP_NODELAY)
    DEFINE_ECONSTANT(IP_TOS)
    DEFINE_ECONSTANT(IP_TTL)
    /*DEFINE_ECONSTANT(IPTOS_LOWDELAY)
    DEFINE_ECONSTANT(IPTOS_THROUGHPUT)
    DEFINE_ECONSTANT(IPTOS_RELIABILITY)
    DEFINE_ECONSTANT(IPTOS_LOWCOST)
    DEFINE_ECONSTANT(IPTOS_MINCOST)*/
    return 0;
}
//=====================================================================================//
CONCEPT_DLL_API ON_DESTROY_CONTEXT MANAGEMENT_PARAMETERS {
#ifdef _WIN32
    WSACleanup();
#endif
	return 0;
}
//=====================================================================================//
CONCEPT_DLL_API CONCEPT_SocketCreate CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK_MIN_MAX(0,1,"SocketCreate: SocketCreate(nTYPE=PROTOCOL_TCP)");

    LOCAL_INIT;
    
    NUMBER type=0;
    int sock=INVALID_SOCKET;

    if (PARAMETERS_COUNT>0) {
        GET_CHECK_NUMBER(0,type,"SocketCreate: parameter 1 should be a number");
    }

    if ((int)type==0) {
        sock=socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    } else
    if ((int)type==1) {
        sock=socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    } else
    if ((int)type==2) {
#ifdef _WIN32
        sock=0x7FFFFFFF;
#else
        sock=socket(PF_UNIX, SOCK_STREAM, 0);
#endif
    } else
    if ((int)type==3) {
#ifdef _WIN32
        sock=-1;
#else
        sock=socket(PF_UNIX, SOCK_DGRAM, 0);
#endif
    }
    int flag=1;
    setsockopt(sock,
                SOL_SOCKET,
                SO_KEEPALIVE,
                (char *)&flag,
                sizeof(flag));

    setsockopt(sock,
                SOL_SOCKET,
                SO_REUSEADDR,
                (char *)&flag,
                sizeof(flag));

    struct linger so_linger;
    so_linger.l_onoff = 1;
    so_linger.l_linger = 10;

    setsockopt(sock,
        SOL_SOCKET,
        SO_LINGER,
        (char *)&so_linger,
        sizeof(so_linger));
#ifndef WIN32
    #ifdef SO_NOSIGPIPE
        if (sock>-1) {
            int flag = 1;
            setsockopt(sock, SOL_SOCKET, SO_NOSIGPIPE, &flag, sizeof(flag));
        }
    #endif
#endif
    RETURN_NUMBER(sock);
    return 0;
}
//=====================================================================================//
CONCEPT_DLL_API CONCEPT_SocketSetOption CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(4,"SocketSetOption: SocketSetOptions(nSocket, nLevel, nOption, nValue)");

    LOCAL_INIT;
    
    NUMBER sock=INVALID_SOCKET;
    NUMBER nLevel=0;
    NUMBER nOption=0;
    NUMBER nValue=0;

    GET_CHECK_NUMBER(0,sock,"SocketSetOption: parameter 1 should be a number");
    GET_CHECK_NUMBER(1,nLevel,"SocketSetOption: parameter 2 should be a number");
    GET_CHECK_NUMBER(2,nOption,"SocketSetOption: parameter 3 should be a number");
    GET_CHECK_NUMBER(3,nValue,"SocketSetOption: parameter 4 should be a number");

#ifdef _WIN32
    if (UnixSocketEmulation[(SYS_INT)sock]) {
        RETURN_NUMBER(-1);
    } else {
        RETURN_NUMBER(setsockopt((int)sock,(int)nLevel,(int)nOption,(char *)(int)nValue, sizeof(int)));
    }
#else
    RETURN_NUMBER(setsockopt((int)sock,(int)nLevel,(int)nOption,(void *)(int)nValue, sizeof(int)));
#endif

    return 0;
}
//=====================================================================================//
CONCEPT_DLL_API CONCEPT_SocketGetOption CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(4,"SocketGetOption: SocketSetOptions(nSocket, nLevel, nOption, var nValue)");

    LOCAL_INIT;
    
    NUMBER sock=INVALID_SOCKET;
    NUMBER nLevel=0;
    NUMBER nOption=0;
    int nValue=0;

    GET_CHECK_NUMBER(0,sock,"SocketGetOption: parameter 1 should be a number");
    GET_CHECK_NUMBER(1,nLevel,"SocketGetOption: parameter 2 should be a number");
    GET_CHECK_NUMBER(2,nOption,"SocketGetOption: parameter 3 should be a number");
    //GET_CHECK_NUMBER(3,nValue,"SocketSetOption: parameter 4 should be a number");



#ifdef _WIN32
    int len = sizeof(nValue);
    if (UnixSocketEmulation[(SYS_INT)sock]) {
        RETURN_NUMBER(-1);
    } else {
        RETURN_NUMBER(getsockopt((int)sock,(int)nLevel,(int)nOption,(char *)(int)nValue, &len));
    }
#else
    socklen_t len = sizeof(nValue);
    RETURN_NUMBER(getsockopt((int)sock,(int)nLevel,(int)nOption,(void *)(int)nValue, &len));
#endif

    SET_NUMBER(3,(int)len);

    return 0;
}
//=====================================================================================//
CONCEPT_DLL_API CONCEPT_SocketError CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(1,"SocketError: SocketError(nSocket)");

    LOCAL_INIT;
    
    NUMBER sock=INVALID_SOCKET;

    GET_CHECK_NUMBER(0,sock,"SocketError: parameter 1 should be a number");

    int err=0;
#ifdef _WIN32
    int res=0;
    if (UnixSocketEmulation[(SYS_INT)sock]) {
        int err = GetLastError();
    } else {
        int errlen = sizeof(err);
        int res=getsockopt((int)sock, SOL_SOCKET, SO_ERROR, (char *)&err, &errlen);
    }
#else
    socklen_t errlen = sizeof(err);
    int res=getsockopt((int)sock, SOL_SOCKET, SO_ERROR, &err, &errlen);
#endif
    if (res<0)
        err=-2;

    RETURN_NUMBER(err);

    return 0;
}
//=====================================================================================//
CONCEPT_DLL_API CONCEPT_SocketConnect CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(3,"SocketConnect: SocketConnect(nSocket, szHostname, nPort)");

    LOCAL_INIT;
    
    NUMBER sock=INVALID_SOCKET;
    char *hostname=0;
    NUMBER nPort=0;

    GET_CHECK_NUMBER(0,sock,"SocketConnect: parameter 1 should be a number");
    GET_CHECK_STRING(1,hostname,"SocketConnect: parameter 2 should be a string");
    GET_CHECK_NUMBER(2,nPort,"SocketConnect: parameter 3 should be a number");


    if (nPort>=0) {
        struct sockaddr_in      sin;
        struct hostent          *hp;
        if ((hp=gethostbyname(hostname))==0) {
            RETURN_NUMBER(-2);
            return 0;
        }
        memset(&sin,0,sizeof(sin));
                                      //INADDR_ANY;  
        sin.sin_addr.s_addr         = ((struct in_addr *)(hp->h_addr))->s_addr;
        sin.sin_family              = AF_INET;
        sin.sin_port                = htons((int)nPort);
        RETURN_NUMBER(connect((int)sock,(struct sockaddr *)  &sin, sizeof(sin)));
    } else {
#ifdef _WIN32
        //RETURN_NUMBER(-1);
        SECURITY_ATTRIBUTES sa = {sizeof(SECURITY_ATTRIBUTES), NULL,TRUE};
        AnsiString pipename("\\\\.\\pipe\\");
        pipename+=hostname;
        HANDLE hpipe = CreateFile(pipename.c_str(), GENERIC_READ | GENERIC_WRITE | SYNCHRONIZE, 0, NULL, OPEN_EXISTING, 0, &sa);
        if (hpipe != INVALID_HANDLE_VALUE) {
            int hpipe_Fd=_open_osfhandle((intptr_t)hpipe, 0);
            UnixSocketEmulation[(SYS_INT)hpipe_Fd]=1;
            SetVariable(PARAMETER(0),VARIABLE_NUMBER,(char *)0,(NUMBER)(SYS_INT)hpipe_Fd);
            RETURN_NUMBER(0);
        } else {
            RETURN_NUMBER(-1);
        }
#else
        struct sockaddr_un address;
        memset(&address, 0, sizeof(struct sockaddr_un));
        address.sun_family = AF_UNIX;
        snprintf(address.sun_path, UNIX_PATH_MAX, hostname);

        RETURN_NUMBER(connect((int)sock,(struct sockaddr *)  &address, sizeof(address)));
#endif
    }
    return 0;
}
//=====================================================================================//
CONCEPT_DLL_API CONCEPT_SocketListen CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK_MIN_MAX(2,4,"SocketListen: SocketListen(nSocket, nPort, nMaxConnections=0x100, interface=\"\")");

    LOCAL_INIT;
    
    NUMBER sock=INVALID_SOCKET;
    NUMBER nMaxCon=0x100;
    NUMBER nPort=0;

    GET_CHECK_NUMBER(0,sock,"SocketListen: parameter 1 should be a number");
    GET_CHECK_NUMBER(1,nPort,"SocketListen: parameter 2 should be a number");

    if (PARAMETERS_COUNT>2) {
        GET_CHECK_NUMBER(2,nMaxCon,"SocketListen: parameter 3 should be a number");
    }

    char *iface=NULL;
    if (PARAMETERS_COUNT>3) {
        GET_CHECK_STRING(3,iface,"SocketListen: parameter 4 should be a string");
    }

    if (nPort>=0) {
        struct sockaddr_in sin;
        memset(&sin, 0, sizeof(sin)); 

        if ((iface) && (iface[0]))
            sin.sin_addr.s_addr         = inet_addr(iface);
        else
            sin.sin_addr.s_addr         = INADDR_ANY; 
        sin.sin_family              = AF_INET;
        sin.sin_port                = htons((int)nPort);

        if (bind((int)sock,(struct sockaddr *)  &sin, sizeof(sin)) == INVALID_SOCKET) {
            RETURN_NUMBER((NUMBER)INVALID_SOCKET);
        }

        RETURN_NUMBER(listen((int)sock,(int)nMaxCon));
    } else {
#ifdef _WIN32
        AnsiString pipename("\\\\.\\pipe\\");
        pipename+=iface;
        if (nMaxCon>0xFF)
            nMaxCon=0xFF;
        HANDLE hpipe = CreateNamedPipe(pipename.c_str(), PIPE_ACCESS_DUPLEX, PIPE_WAIT, (int)nMaxCon, 0x3FFF, 0x3FFF, 10000, NULL);
        if (hpipe != INVALID_HANDLE_VALUE) {
            int hpipe_Fd=_open_osfhandle((intptr_t)hpipe, 0);
            UnixSocketEmulation[(SYS_INT)hpipe_Fd]=1;
            SetVariable(PARAMETER(0),VARIABLE_NUMBER,(char *)0,(NUMBER)(SYS_INT)hpipe_Fd);
            RETURN_NUMBER(0);
        } else {
            RETURN_NUMBER(-1);
        }
#else
        struct sockaddr_un address;
        memset(&address, 0, sizeof(struct sockaddr_un));
        address.sun_family = AF_UNIX;

        snprintf(address.sun_path, UNIX_PATH_MAX, iface);

        if (bind((int)sock,(struct sockaddr *)  &address, sizeof(address)) == INVALID_SOCKET) {
            RETURN_NUMBER((NUMBER)INVALID_SOCKET);
        }
        RETURN_NUMBER(listen((int)sock,(int)nMaxCon));
#endif
    }
    return 0;
}
//=====================================================================================//
CONCEPT_DLL_API CONCEPT_SocketBindUDP CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK_MIN_MAX(2,3,"SocketBindUDP: SocketBindUDP(nSocket, nPort, interface=\"\")");

    LOCAL_INIT;
    
    NUMBER sock=INVALID_SOCKET;
    NUMBER nPort=0;

    GET_CHECK_NUMBER(0,sock,"SocketBindUDP: parameter 1 should be a number");
    GET_CHECK_NUMBER(1,nPort,"SocketBindUDP: parameter 2 should be a number");

    char *iface=NULL;
    if (PARAMETERS_COUNT>2) {
        GET_CHECK_STRING(2,iface,"SocketBindUDP: parameter 3 should be a string");
    }

    struct sockaddr_in sin;
    memset(&sin, 0, sizeof(sin)); 
    if ((iface) && (iface[0]))
        sin.sin_addr.s_addr         = inet_addr(iface);
    else
        sin.sin_addr.s_addr         = INADDR_ANY;  
    sin.sin_family              = AF_INET;
    sin.sin_port                = htons((int)nPort);

    int res=bind((int)sock,(struct sockaddr *)  &sin, sizeof(sin));

    RETURN_NUMBER(res);

    return 0;
}
//=====================================================================================//
CONCEPT_DLL_API CONCEPT_SocketAccept CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(1,"SocketAccept: SocketAccept(nSocket)");

    LOCAL_INIT;
    
    NUMBER sock=INVALID_SOCKET;
    NUMBER nPort=0;

    GET_CHECK_NUMBER(0,sock,"SocketAccept: parameter 1 should be a number");

#ifdef _WIN32
    int ref_count=UnixSocketEmulation[(SYS_INT)sock];
    if (ref_count) {
        ref_count++;
        if (ConnectNamedPipe((HANDLE)_get_osfhandle((SYS_INT)sock), NULL)) {
            UnixSocketEmulation[(SYS_INT)sock]=ref_count;
            RETURN_NUMBER(sock);
        } else {
            RETURN_NUMBER(-1);
        }
        return 0;
    }
#endif
    RETURN_NUMBER(accept((int)sock, NULL, NULL));
    return 0;
}
//=====================================================================================//
CONCEPT_DLL_API CONCEPT_SocketClose CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(1,"SocketClose: SocketClose(nSocket)");

    LOCAL_INIT;
    
    NUMBER sock=INVALID_SOCKET;

    GET_CHECK_NUMBER(0,sock,"SocketClose: parameter 1 should be a number");

#ifdef _WIN32
    int ref_count=UnixSocketEmulation[(SYS_INT)sock];
    if (ref_count) {
        ref_count--;
        if (ref_count>0) {
            UnixSocketEmulation[(SYS_INT)sock]=ref_count;
        } else {
            UnixSocketEmulation.erase((SYS_INT)sock);
            //CloseHandle((HANDLE)(SYS_INT)sock);
            close((int)sock);
        }
    } else {
        shutdown((int)sock, SD_BOTH);
        RETURN_NUMBER(closesocket((int)sock));
    }
#else
    shutdown((int)sock, SHUT_RDWR);
    RETURN_NUMBER(close((int)sock));
#endif


    return 0;
}
//=====================================================================================//
CONCEPT_DLL_API CONCEPT_SocketRead CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK_MIN_MAX(3,7,"SocketRead: SocketRead(nSocket, var anyBuffer, nSize[, bSocketIsUDP=false[, bReceiveExactSize=false][, var ip, var port]])");

    LOCAL_INIT;
    
    NUMBER sock=INVALID_SOCKET;
    NUMBER nSize=0;
    NUMBER isUdp=0;
    NUMBER exactSize=0;

    GET_CHECK_NUMBER(0,sock,"SocketRead: parameter 1 should be a number");
    GET_CHECK_NUMBER(2,nSize,"SocketRead: parameter 3 should be a number");

    if (PARAMETERS_COUNT>3) {
        GET_CHECK_NUMBER(3,isUdp,"SocketRead: parameter 4 should be a number");

        if (PARAMETERS_COUNT>4) {
            GET_CHECK_NUMBER(4,exactSize,"SocketRead: parameter 5 should be a number");
        }
    }

    if ((int)nSize<=0) {
        RETURN_NUMBER(0);
        return 0;
    }

    //char *buf=new char[(int)nSize+1];
    char *buf=0;
    CORE_NEW((int)nSize+1, buf);
    if (!buf) {
        RETURN_NUMBER(-3);
        return 0;
    }
    buf[0]=0;
    buf[(int)nSize]=0;
#ifdef _WIN32
    int ref_count=UnixSocketEmulation[(SYS_INT)sock];
    if (ref_count) {
        DWORD outsize=0;
        if (ReadFile((HANDLE)_get_osfhandle((SYS_INT)sock), buf, (int)nSize, &outsize, NULL)) {
            SetVariable(PARAMETER(1),-1,buf,outsize);
            RETURN_NUMBER(outsize);
        } else {
            CORE_DELETE(buf);
            SET_STRING(1, "");
            RETURN_NUMBER(-1);
        }
        return 0;
    }
#endif
    int rec_size=0;
    if (isUdp) {
    #ifdef _WIN32
        struct sockaddr host_address;
	    int addr_length =   sizeof(host_address);
        rec_size=recvfrom((int)sock, buf, (int)nSize,0,&host_address,&addr_length);
    #else
        struct sockaddr host_address;
	    socklen_t addr_length =   sizeof(host_address);
        rec_size=recvfrom((int)sock, buf, (int)nSize,0,&host_address,&addr_length);
    #endif
        if (PARAMETERS_COUNT>5) {
            char *ip=0;
            int port=0;
            char  ipstr[INET6_ADDRSTRLEN];
            if (host_address.sa_family == AF_INET) {
                struct sockaddr_in *s = (struct sockaddr_in *)&host_address;
                ip   = (char *)inet_ntop(AF_INET, &s->sin_addr, ipstr, sizeof ipstr);
                port = ntohs(s->sin_port);
            } else
            if (host_address.sa_family == AF_INET6) {
                struct sockaddr_in6 *s = (struct sockaddr_in6 *)&host_address;
                ip   = (char *)inet_ntop(AF_INET6, &s->sin6_addr, ipstr, sizeof ipstr);
                port = ntohs(s->sin6_port);
            }
            if (ip) {
                SET_STRING(5, ip);
            } else {
                SET_STRING(5, "");
            }
            if (PARAMETERS_COUNT>6) {
                SET_NUMBER(6, port);
            }
        }
        
    } else {
        int len=0;
        do {
            len=recv((int)sock, buf+rec_size, (int)nSize-rec_size,0);

            if (len>0)
                rec_size+=len;

        } while (((int)exactSize) && (len>0) && (rec_size<(int)nSize));
    }
    if (rec_size>0) {
        SetVariable(PARAMETER(1),-1,buf,rec_size);
    } else {
        CORE_DELETE(buf);
        SET_STRING(1, "");
    }
    //SET_BUFFER(1, buf, rec_size);

    //delete buf;

    RETURN_NUMBER(rec_size);

    return 0;
}
//=====================================================================================//
CONCEPT_DLL_API CONCEPT_SocketWrite CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK_MIN_MAX(2,5,"SocketWrite: SocketWrite(nSocket, anyBuffer[, bSocketIsUDP=false, szUDPHost, nUDPPort])");

    LOCAL_INIT;
    
    NUMBER sock=INVALID_SOCKET;
    char *buffer;
    NUMBER isUdp=0;
    NUMBER nSize=0;
    NUMBER nPort=0;
    char *hostname=0;

    GET_CHECK_NUMBER(0,sock,"SocketWrite: parameter 1 should be a number");
    GET_CHECK_BUFFER(1,buffer,nSize,"SocketWrite: parameter 2 should be a string buffer");

    struct hostent          *hp;
    if (PARAMETERS_COUNT==5) {
        GET_CHECK_NUMBER(2,isUdp,"SocketWrite: parameter 3 should be a number");

        GET_CHECK_STRING(3,hostname,"SocketWirte: parameter 4 should be a string");
        GET_CHECK_NUMBER(4,nPort,"SocketWrite: parameter 5 should be a number");

        if ((hp=gethostbyname(hostname))==0) {
            RETURN_NUMBER(-2);
            return 0;
        }
    }
#ifdef _WIN32
    int ref_count=UnixSocketEmulation[(SYS_INT)sock];
    if (ref_count) {
        DWORD outsize=0;
        if (WriteFile((HANDLE)_get_osfhandle((SYS_INT)sock), buffer, (int)nSize, &outsize, NULL)) {
            RETURN_NUMBER(outsize);
        } else {
            RETURN_NUMBER(-1);
        }
        return 0;
    }
#endif
    int sent_size=0;
    if (isUdp) {
    #ifdef _WIN32
        struct sockaddr_in host_address;
	    int addr_length =   sizeof(host_address);
    #else
        struct sockaddr_in host_address;
	    socklen_t addr_length =   sizeof(host_address);
    #endif
        host_address.sin_family=AF_INET;
        host_address.sin_port=htons((int)nPort);
        host_address.sin_addr.s_addr         = ((struct in_addr *)(hp->h_addr))->s_addr;


        sent_size=sendto((int)sock, buffer, (int)nSize,0,(struct sockaddr *)&host_address,addr_length);
    } else {
        sent_size=send((int)sock, buffer, (int)nSize,0);
    }

    RETURN_NUMBER(sent_size);

    return 0;
}
//=====================================================================================//
CONCEPT_DLL_API CONCEPT_SocketInfo CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(1,"SocketInfo: SocketInfo(nSocket)");

    LOCAL_INIT;
    
    NUMBER sock=INVALID_SOCKET;

    GET_CHECK_NUMBER(0,sock,"SocketInfo: parameter 1 should be a number");


    /*struct sockaddr_in peer;
#ifdef _WIN32
    int peerlen = sizeof(struct sockaddr);
#else
    socklen_t peerlen = sizeof(struct sockaddr);
#endif
    getpeername((int)sock,(sockaddr *)&peer,&peerlen);*/

    struct sockaddr_storage addr;
    char        ipstr[INET6_ADDRSTRLEN];
    int         port     = 0;
    static char *unknown = "unknown";

#ifdef _WIN32
    int peerlen = sizeof(addr);
#else
    socklen_t peerlen = sizeof(addr);
#endif
    Invoke(INVOKE_CREATE_ARRAY, RESULT, 0);
#ifdef _WIN32
    int ref_count=UnixSocketEmulation[(SYS_INT)sock];
    if (ref_count)
        return 0;
#endif
    if (getpeername((int)sock, (sockaddr *)&addr, &peerlen))
        return 0;

    const char *ip = 0;
    if (addr.ss_family == AF_INET) {
        struct sockaddr_in *s = (struct sockaddr_in *)&addr;
        port = ntohs(s->sin_port);
        ip   = inet_ntop(AF_INET, &s->sin_addr, ipstr, sizeof ipstr);
    } else
    if (addr.ss_family == AF_INET6) {
        struct sockaddr_in6 *s = (struct sockaddr_in6 *)&addr;
        port = ntohs(s->sin6_port);
        ip   = inet_ntop(AF_INET6, &s->sin6_addr, ipstr, sizeof ipstr);
    }
    if (!ip)
        ip=unknown;


    Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY,RESULT,"address",(INTEGER)VARIABLE_STRING,(char *)ip,(NUMBER)0);
    Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY,RESULT,"port",(INTEGER)VARIABLE_NUMBER,(char *)"",(NUMBER)port);

    return 0;
}
//=====================================================================================//
CONCEPT_DLL_API CONCEPT_SocketHasData CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(1,"SocketHasData: SocketHasData(nSocket)");

    LOCAL_INIT;
    
    NUMBER sock=INVALID_SOCKET;

    int has_data=0;
    GET_CHECK_NUMBER(0,sock,"SocketHasData: parameter 1 should be a number");

#ifdef _WIN32
    int ref_count=UnixSocketEmulation[(SYS_INT)sock];
    if (ref_count) {
        DWORD has_bytes=0;
        if (PeekNamedPipe((HANDLE)_get_osfhandle((SYS_INT)sock), NULL, 0, NULL, &has_bytes, NULL)) {
            if (has_bytes) {
                RETURN_NUMBER(1);
            } else {
                RETURN_NUMBER(0);
            }
        } else {
            RETURN_NUMBER(-1);
        }
        return 0;
    }
#endif
    struct timeval timeout;
    timeout.tv_sec = 0;
    timeout.tv_usec = 100;
    fd_set socks;

    FD_ZERO(&socks);
    FD_SET((int)sock,&socks);

    int sel_val=select(FD_SETSIZE, &socks, 0, 0, &timeout);

    has_data=(sel_val!=0);
    if (sel_val<0) {
        RETURN_NUMBER(-1);
    } else {
        RETURN_NUMBER(has_data);
    }
    return 0;
}
//=====================================================================================//
CONCEPT_FUNCTION_IMPL(WSParseHandshake, 1)
    T_STRING(0)
    struct handshake hs;
    ws_frame_type frame_type = ws_parse_handshake((uint8_t *)PARAM(0), PARAM_LEN(0), &hs);
    CREATE_ARRAY(RESULT);
    Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY,RESULT,"frame_type",(INTEGER)VARIABLE_NUMBER, 0, (NUMBER)frame_type);
    if (frame_type!=WS_INCOMPLETE_FRAME) {
        Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY,RESULT,"resource",(INTEGER)VARIABLE_STRING, hs.resource ? hs.resource : "", (NUMBER)0);
        Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY,RESULT,"host",(INTEGER)VARIABLE_STRING, hs.host ? hs.host : "", (NUMBER)0);
        Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY,RESULT,"origin",(INTEGER)VARIABLE_STRING, hs.origin ? hs.origin : "", (NUMBER)0);
        Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY,RESULT,"protocol",(INTEGER)VARIABLE_STRING, hs.protocol ? hs.protocol : "", (NUMBER)0);
        Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY,RESULT,"version",(INTEGER)VARIABLE_NUMBER, "", (NUMBER)hs.version);
        Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY,RESULT,"key",(INTEGER)VARIABLE_STRING, hs.key ? hs.key : "", (NUMBER)0);
        Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY,RESULT,"key1",(INTEGER)VARIABLE_STRING, hs.key1 ? hs.key1 : "", (NUMBER)0);
        Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY,RESULT,"key2",(INTEGER)VARIABLE_STRING, hs.key2 ? hs.key2 : "", (NUMBER)0);
        Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY,RESULT,"key3",(INTEGER)VARIABLE_STRING, hs.key3, (NUMBER)8);
    }
END_IMPL
//=====================================================================================//
CONCEPT_FUNCTION_IMPL(WSHandshakeAnswer, 1)
    T_ARRAY(0)
    struct handshake hs;
    nullhandshake(&hs);
    INTEGER type=0;
    NUMBER dummy=0;
    char *str=0;
    uint8_t buffer[0xFFFF];
    size_t out_len=0xFFFF;

    Invoke(INVOKE_GET_ARRAY_ELEMENT_BY_KEY, PARAMETER(0), "resource", &type, &str, &dummy);
    if (type==VARIABLE_STRING)
        hs.resource=str;
    Invoke(INVOKE_GET_ARRAY_ELEMENT_BY_KEY, PARAMETER(0), "host", &type, &str, &dummy);
    if (type==VARIABLE_STRING)
        hs.host=str;
    Invoke(INVOKE_GET_ARRAY_ELEMENT_BY_KEY, PARAMETER(0), "origin", &type, &str, &dummy);
    if (type==VARIABLE_STRING)
        hs.origin=str;
    Invoke(INVOKE_GET_ARRAY_ELEMENT_BY_KEY, PARAMETER(0), "protocol", &type, &str, &dummy);
    if (type==VARIABLE_STRING)
        hs.protocol=str;
    Invoke(INVOKE_GET_ARRAY_ELEMENT_BY_KEY, PARAMETER(0), "version", &type, &str, &dummy);
    if (type==VARIABLE_NUMBER)
        hs.version=(char)dummy;
    Invoke(INVOKE_GET_ARRAY_ELEMENT_BY_KEY, PARAMETER(0), "key", &type, &str, &dummy);
    if (type==VARIABLE_STRING)
        hs.key=str;
    Invoke(INVOKE_GET_ARRAY_ELEMENT_BY_KEY, PARAMETER(0), "key1", &type, &str, &dummy);
    if (type==VARIABLE_STRING)
        hs.key1=str;
    Invoke(INVOKE_GET_ARRAY_ELEMENT_BY_KEY, PARAMETER(0), "key2", &type, &str, &dummy);
    if (type==VARIABLE_STRING)
        hs.key2=str;
    Invoke(INVOKE_GET_ARRAY_ELEMENT_BY_KEY, PARAMETER(0), "key3", &type, &str, &dummy);
    if ((type==VARIABLE_STRING) && (str)) {
        int len=(int)dummy;
        if (len>8)
            len=8;
        memcpy(hs.key3, str, len);
    }
    if ((!hs.origin) || (!hs.host) || (!hs.resource)) {
        RETURN_STRING("");
        return 0;
    }
    ws_get_handshake_answer(&hs, buffer, &out_len);
    if (out_len>0) {
        RETURN_BUFFER((char *)buffer, out_len);
    } else {
        RETURN_STRING("");
    }
END_IMPL
//=====================================================================================//
CONCEPT_FUNCTION_IMPL_MINMAX_PARAMS(WSMakeFrame, 1, 2)
    T_STRING(0)

    ws_frame_type ft_orig=WS_TEXT_FRAME;
    if (PARAMETERS_COUNT>1) {
        T_NUMBER(1)
        ft_orig=(ws_frame_type)PARAM_INT(1);
    }
    char *buffer=0;
    size_t out_len=PARAM_LEN(0)+0xFF;
    CORE_NEW(out_len+1, buffer);

    ws_frame_type frame_type=ws_make_frame((uint8_t *)PARAM(0), PARAM_LEN(0), (uint8_t *)buffer, &out_len, ft_orig);
    if ((frame_type!=ft_orig) || (!out_len)) {
        CORE_DELETE(buffer);
        RETURN_STRING("");
    } else
        SetVariable(RESULT,-1,buffer,out_len);
END_IMPL
//=====================================================================================//
CONCEPT_FUNCTION_IMPL_MINMAX_PARAMS(WSParseFrame, 1, 2)
    T_STRING(0)

    ws_frame_type ft_orig=WS_BINARY_FRAME;
    char *buffer=0;
    size_t out_len=0;

    ws_frame_type frame_type=ws_parse_input_frame((uint8_t *)PARAM(0), PARAM_LEN(0), (uint8_t **)&buffer, &out_len);
    if ((out_len) && (buffer)) {
        RETURN_BUFFER(buffer, out_len);
    } else {
        RETURN_STRING("");
    }
    if (PARAMETERS_COUNT>1) {
        SET_NUMBER(1, frame_type);
    }
    
END_IMPL
//=====================================================================================//
CONCEPT_FUNCTION_IMPL(GetHostName, 0)
    char buffer[0xFFF];
    buffer[0]=0;
    int res=gethostname(buffer, 0xFFE);
    if (res) {
        RETURN_STRING("");
    } else {
        RETURN_STRING(buffer);
    }
END_IMPL
//=====================================================================================//
CONCEPT_FUNCTION_IMPL(GetHostByAddr, 1)
    T_STRING(0);
    struct hostent *he;
#ifdef _WIN32
    struct in_addr addr = { 0 };

    addr.s_addr = inet_addr(PARAM(0));
    he = gethostbyaddr((char *) &addr, 4, AF_INET);

    if (he) {
        RETURN_STRING(he->h_name);
    } else {
        RETURN_STRING("");
    }
#else
    struct in_addr ipv4addr;

    inet_pton(AF_INET, PARAM(0), &ipv4addr);
    he = gethostbyaddr((char *)&ipv4addr, sizeof ipv4addr, AF_INET);

    if (he) {
        RETURN_STRING(he->h_name);
    } else {
        RETURN_STRING("");
    }
#endif
END_IMPL
//=====================================================================================//
CONCEPT_FUNCTION_IMPL(GetHostByName, 1)
    T_STRING(0);
    struct hostent *he;
    he = gethostbyname(PARAM(0));
#ifdef _WIN32
    struct in_addr addr = { 0 };
#else
    struct in_addr addr;
#endif

    Invoke(INVOKE_CREATE_ARRAY, RESULT, 0);
    if (he) {
        if (he->h_name)
            Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY,RESULT,"h_name",(INTEGER)VARIABLE_STRING,(char *)he->h_name,(NUMBER)0);
        else
            Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY,RESULT,"h_name",(INTEGER)VARIABLE_STRING,(char *)"",(NUMBER)0);

        int idx=0;
        for (char **pAlias = he->h_aliases; *pAlias != 0; pAlias++) {
                    AnsiString h_aliases("h_aliases");
                    h_aliases+=AnsiString((long)idx);
                    Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY,RESULT,h_aliases.c_str(),(INTEGER)VARIABLE_STRING,(char *)pAlias,(NUMBER)0);
        }

        if (he->h_addrtype == AF_INET) {
            int i=0;
            while (he->h_addr_list[i] != 0) {
                addr.s_addr = *(u_long *) he->h_addr_list[i];
                AnsiString h_aliases("h_addr_list");
                h_aliases+=AnsiString((long)i);
                Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY,RESULT,h_aliases.c_str(),(INTEGER)VARIABLE_STRING,inet_ntoa(addr),(NUMBER)0);
                i++;
            }
        }
    }
END_IMPL
//=====================================================================================//
CONCEPT_FUNCTION_IMPL_MINMAX_PARAMS(Ping, 1, 5)
    T_STRING(0);
    // count, timeout, message size (last 2 reversed);
    int nCount = 4;
    int nMessageSize = 32;
    int nTimeOut = 5000;

    if (PARAMETERS_COUNT>2) {
        T_NUMBER(2)
        nCount=PARAM_INT(2);
    }
    if (PARAMETERS_COUNT>3) {
        T_NUMBER(3)
        nTimeOut=PARAM_INT(3);
    }
    if (PARAMETERS_COUNT>4) {
        T_NUMBER(4)
        nMessageSize=PARAM_INT(4);
    }
    CREATE_ARRAY(RESULT);
    int res=Ping(PARAM(0), nCount, nMessageSize, nTimeOut, Invoke, RESULT);
    //RETURN_NUMBER(res);
    if (PARAMETERS_COUNT>1) {
        SET_NUMBER(1, res);
    }
END_IMPL
//=====================================================================================//
CONCEPT_FUNCTION_IMPL(htonl, 1)
    T_NUMBER(0)
    int res=htonl(PARAM_INT(0));
    RETURN_NUMBER(res);
END_IMPL
//=====================================================================================//
CONCEPT_FUNCTION_IMPL(htons, 1)
    T_NUMBER(0)
    int res=htons(PARAM_INT(0));
    RETURN_NUMBER(res);
END_IMPL
//=====================================================================================//
CONCEPT_FUNCTION_IMPL(ntohl, 1)
    T_NUMBER(0)
    int res=ntohl(PARAM_INT(0));
    RETURN_NUMBER(res);
END_IMPL
//=====================================================================================//
CONCEPT_FUNCTION_IMPL(ntohs, 1)
    T_NUMBER(0)
    int res=ntohs(PARAM_INT(0));
    RETURN_NUMBER(res);
END_IMPL
//=====================================================================================//
CONCEPT_FUNCTION_IMPL(ListInterfaces, 0)
    std::vector <DataHolder> holder;
    CREATE_ARRAY(RESULT);
    if (getInterfaces(&holder)==0) {
        INTEGER idx=0;
        for (std::vector<DataHolder>::const_iterator i = holder.begin (); i != holder.end (); ++i) {
            DataHolder d = *i;
            if (d.ip.size()) {
                if (!d.mac.size()) {
                    for (std::vector<DataHolder>::const_iterator i2 = holder.begin (); i2 != holder.end (); ++i2) {
                        DataHolder d2 = *i2;
                        if ((d2.adapter==d.adapter) && (d2.mac.size())) {
                            d.mac=d2.mac;
                        }
                    }
                }
                void *elem = 0;
                Invoke(INVOKE_ARRAY_VARIABLE,RESULT,idx,&elem);
                idx++;
                if (elem) {
                    CREATE_ARRAY(elem);

                    Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY, elem, "ip", (INTEGER)VARIABLE_STRING, (char *)d.ip.c_str(), (NUMBER)d.ip.size());
                    Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY, elem, "mask", (INTEGER)VARIABLE_STRING, (char *)d.mask.c_str(), (NUMBER)d.mask.size());
                    Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY, elem, "mac", (INTEGER)VARIABLE_STRING, (char *)d.mac.c_str(), (NUMBER)d.mac.size());
                    Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY, elem, "gateway", (INTEGER)VARIABLE_STRING, (char *)d.gateway.c_str(), (NUMBER)d.gateway.size());
                    Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY, elem, "adapter", (INTEGER)VARIABLE_STRING, (char *)d.adapter.c_str(), (NUMBER)d.adapter.size());
                    Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY, elem, "description", (INTEGER)VARIABLE_STRING, (char *)d.description.c_str(), (NUMBER)d.description.size());
                    Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY, elem, "flags", (INTEGER)VARIABLE_NUMBER, (char *)0, (NUMBER)d.flags);
                    Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY, elem, "type", (INTEGER)VARIABLE_NUMBER, (char *)0, (NUMBER)d.type);
                }
            }
        }
    }
END_IMPL
//=====================================================================================//
