//------------ standard header -----------------------------------//
#include "stdlibrary.h"
//------------ end of standard header ----------------------------//
#include "library.h"
#include "AnsiString.h"
#include "AnsiList.h"
//#include "xmlParser.h"
#include <time.h>
#include <string.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdint.h>
#include "semhh.h"

struct WriteNode {
    char *buffer;
    int size;
    void *Next;
};

struct TimerData {
    long RID;
    int repeat;
    int APID;
    unsigned int ms;
};

#define CACHE_SIZE      0xFFFF
#define CHUNK_SIZE      0x80000

#ifdef _WIN32
    #include <io.h>
    #include <winsock2.h>
    #include <ws2tcpip.h>
#else
    #include <unistd.h>
    #include <sys/types.h>
    #include <sys/socket.h>
    #include <netinet/in.h>
    #include <netinet/tcp.h>
    #include <arpa/inet.h>
    #include <poll.h>
    #include <sys/select.h>
    #include <time.h> // sleep

    int eof(int stream) {
        struct timeval timeout;
        timeout.tv_sec = 0;
        timeout.tv_usec = 10;
        fd_set s;
        FD_ZERO(&s);
        FD_SET(stream,&s);

        int val=select(stream+1, &s, 0, 0, &timeout);

        if (val) {
           if (FD_ISSET(stream, &s))
               return 0;
        }
		return 1;
    }

    int is_writeable(int fd) {
        struct pollfd p;
        int ret;

        p.fd = fd;
        p.events = POLLOUT;

        ret = poll(&p, 1, 0);

        if (ret < 0)
            return 0;

        return p.revents & POLLOUT;
    }

#endif

#define SOCKET_RECREATE_CHECK                                                                               \
    if (event_timeout > 0) {                                                                                \
        int new_socket=Invoke(INVOKE_GET_PROTO_SOCKET, MAIN_HANDLER);                           \
        if (CLIENT_SOCKET!=new_socket) {                                                        \
            send_failed=0;                                                                      \
            last_msg_seconds=time(NULL);                                                        \
            SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],VARIABLE_NUMBER,"",0);      \
            SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[2]-1],VARIABLE_STRING,"",0);      \
            SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[3]-1],VARIABLE_STRING,"",0);      \
            SetVariable(RESULT,VARIABLE_NUMBER,"",1);                                           \
            ResetConceptPeer(new_socket);                                                       \
            return 0;                                                                           \
        }                                                                                       \
    }


#define SOCKET_RECREATE_CHECK_WAIT_MESSAGE                                                  \
    int new_socket=Invoke(INVOKE_GET_PROTO_SOCKET, MAIN_HANDLER);                           \
    if (CLIENT_SOCKET!=new_socket) {                                                        \
        last_msg_seconds=time(NULL);                                                        \
        send_failed=0;                                                                      \
        SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],VARIABLE_NUMBER,"",0);      \
        SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[2]-1],VARIABLE_STRING,"",0);      \
        SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[3]-1],VARIABLE_STRING,"",0);      \
        SetVariable(RESULT,VARIABLE_NUMBER,"",1);                                           \
        ResetConceptPeer(new_socket);                                                       \
        return 0;                                                                           \
    }                                                                                       \

#include "INTERFACE_TO_AES.h"

AnsiList  BufferedMessages;
AnsiList  ThreadedMessages;

#ifdef _WIN32
	#define pipe(phandles)	_pipe (phandles, 4096, _O_BINARY)
#endif

char has_debug=0;
int DebugPipes[2];

static int pipe_in=0;
static int pipe_out=0;
static int apid=0;
static int parent_apid=0;

int is_cached=0;
char *cached_buffer=0;
int cached_size=0;
int alloc_cached_size=0;
void *MAIN_HANDLER=0;
static int RTSOCKET=-1;
static struct sockaddr_storage remote_conceptaddr;
static struct sockaddr_storage remote_conceptudpaddr;
static socklen_t remote_len=0;
static char RTCONFIRMED=0;

static HHSEM sem_send;
static HHSEM sem_recv;
static HHSEM sem_apwrite;
static HHSEM sem_bufferlist;
static WriteNode *FIRST=0;
static int send_failed=0;
static int event_timeout=0;
static time_t last_msg_seconds=0;

#ifdef _WIN32
    LPVOID       threadID;
#else
    pthread_t    threadID;
#endif

int sock_eof(int stream) {
    struct timeval timeout;
    timeout.tv_sec = 0;
    timeout.tv_usec = 10;
    fd_set s;
    FD_ZERO(&s);
    FD_SET(stream,&s);

    int val=select(stream+1, &s, 0, 0, &timeout);

    if (val<0)
        return 0;
    if (val) {
       if (FD_ISSET(stream, &s))
           return 0;
    }
    return 1;
}


int sock_eof_timeout(int stream, int timeoutms) {
    int sec=timeoutms/1000;
    int ms=timeoutms%1000;

    struct timeval timeout;
    timeout.tv_sec = sec;
    timeout.tv_usec = ms*1000;

    fd_set s;
    FD_ZERO(&s);
    FD_SET(stream,&s);

    int val=select(stream+1, &s, 0, 0, &timeout);

    if (val<0)
        return -1;
    if (val) {
       if (FD_ISSET(stream, &s))
           return 0;
    }
    return 1;
}

void ResetConceptPeer(SYSTEM_SOCKET socket) {
    // disable UDP traffic on network change
    if (RTSOCKET>0) {
#ifdef _WIN32
        closesocket(RTSOCKET);
#else
        close(RTSOCKET);
#endif
        RTSOCKET=-1;
        RTCONFIRMED=0;
    }
}

int deturnated_send(int CLIENT_SOCKET,char *buffer, int size, int extra=0, int is_udp=0) {
    int ssize=0;
    int orig_size=size;
    semp(sem_send);
    do {
        int to_send=size;
        if (size>CHUNK_SIZE)
            to_send=CHUNK_SIZE;
        if ((is_udp) && (RTSOCKET>0) && (RTCONFIRMED) && (remote_conceptudpaddr.ss_family)) {
            ssize=sendto(RTSOCKET, buffer, to_send, 0, (struct sockaddr *)&remote_conceptudpaddr, remote_len);
            if (ssize<=0) {
                ResetConceptPeer(CLIENT_SOCKET);
                continue;
            }
        } else
            ssize=send(CLIENT_SOCKET, buffer, to_send, extra);

        if (ssize<=0) {
#ifdef _WIN32
			if (WSAGetLastError()==WSAEWOULDBLOCK) {
                Sleep(5);
                continue;
            }
#else
			if (errno==EAGAIN) {
                usleep(5000);
                continue;
            }
#endif


            if (ssize<0)
                send_failed=1;
            semv(sem_send);
            return -1;
        }

        buffer+=ssize;
        size-=ssize;

    } while (size);

    semv(sem_send);
    return orig_size;
}
//-----------------------------------------------------------------------------------
/*int Cache(char *buffer,long size) {
    int old_size=cached_size;
    cached_size+=size+sizeof(long);
    cached_buffer=(char *)realloc(cached_buffer,cached_size);
    *(long *)(cached_buffer+old_size)=htonl(size);
    memcpy(cached_buffer+old_size+sizeof(long),buffer,size);
    return 1;
}*/
//-----------------------------------------------------------------------------------
int Cache(char *buffer,int size) {
    int old_size=cached_size;
    cached_size+=size+sizeof(int);
    if (alloc_cached_size<cached_size) {
        alloc_cached_size+=(cached_size/CACHE_SIZE + 1)*CACHE_SIZE;
        cached_buffer=(char *)realloc(cached_buffer,/*cached_size*/alloc_cached_size);
    }
    *(int *)(cached_buffer+old_size)=htonl(size);
    memcpy(cached_buffer+old_size+sizeof(int),buffer,size);
    return 1;
}
//-----------------------------------------------------------------------------------
int FlushCache(int CLIENT_SOCKET) {
    int res=-1;
    if ((cached_size) && (cached_buffer)) {
        res=deturnated_send(CLIENT_SOCKET,cached_buffer, cached_size, 0);
        free(cached_buffer);
        cached_size=0;
        alloc_cached_size=0;
        cached_buffer=0;
    }
    return res;
}
//-----------------------------------------------------------------------------------
static char *empty_char=(char *)"\0";
static char *def_msg="350";
int DeSerializeBuffer2(char *buffer, int size, char **Owner, int *owner_len, int *message, char **Target, int *len_target, char **Value, int *len_value, unsigned int compressed=0, char *buf_owner=0) {
    if (compressed) {
    	compressed &= 0xF0000000;
    	unsigned int owner=compressed >> 16;
    	int val_len=compressed & 0xFFFF;

    	//*Owner=(long)owner;
        // event fired
        if (buf_owner) {
            buf_owner[0]=0;
            sprintf(buf_owner, "%i", owner);
            *Owner=buf_owner;
            *owner_len=strlen(buf_owner);
        }
    	*message=0x1001;
        // on buffer
        *Target=def_msg;
        *len_target=3;
    	*Value=buffer;
        *len_value=val_len;
    	return 1;
    }
    // OWNER poate avea maxim 256 octeti
    char *ptr=buffer;
    unsigned char owner_size=*(unsigned char *)buffer;
    buffer++;
    *owner_len=owner_size;
    if (owner_size) {
        *Owner=buffer;
        buffer+=owner_size;
    } else
        *Owner=empty_char;

    *message=ntohl(*(int *)buffer);
    buffer+=sizeof(int);
    unsigned short target_size=ntohs(*(unsigned short *)buffer);
    buffer+=sizeof(short);
    *len_target=target_size;
    if (target_size) {
        *Target=buffer;
        buffer+=target_size;
    } else
        *Target=empty_char;

    *len_value=size-((uintptr_t)buffer-(uintptr_t)ptr);
    if (*len_value > 0)
        *Value=buffer;
    else {
        *Value=empty_char;
        *len_value=0;
    }
    return 1;
}
//-----------------------------------------------------------------------------------
int DeSerializeBuffer(char *buffer, int size, AnsiString *Owner, int *message, AnsiString *Target, AnsiString *Value, unsigned int compressed=0) {
    if (compressed) {
    	compressed &= 0x0FFFFFFF;
    	int owner=compressed >> 16;
    	int val_len=compressed & 0xFFFF;

    	*Owner=(long)owner;
        // event fired
    	*message=0x1001;
        // on buffer
    	Target->LoadBuffer("350", 3);
    	Value->LoadBuffer(buffer, val_len);
    	return 1;
    }
    // OWNER poate avea maxim 256 octeti
    char *ptr=buffer;
    unsigned char owner_size=*(unsigned char *)buffer;
    buffer++;
    if (owner_size) {
        Owner->LoadBuffer(buffer, owner_size);
        buffer+=owner_size;
    } else
        *Owner=(char *)"";

    *message=ntohl(*(int *)buffer);
    buffer+=sizeof(int);
    unsigned short target_size=ntohs(*(unsigned short *)buffer);
    buffer+=sizeof(short);
    if (target_size) {
        Target->LoadBuffer(buffer, target_size);
        buffer+=target_size;
    } else
        *Target=(char *)"";

    long last_size=size-((uintptr_t)buffer-(uintptr_t)ptr);
    if (last_size>0)
        Value->LoadBuffer(buffer,last_size);
    else
        *Value=(char *)"";
    return 1;
}
//-----------------------------------------------------------------------------------
char *SerializeBuffer(char **buffer, int *size, AnsiString *Owner, int message, AnsiString *Target, AnsiString *Value, char *key, int *is_udp=NULL) {
    // OWNER poate avea maxim 256 octeti
    int vlen=Value->Length();
    int tlen=Target->Length();
    if ((message==0x110) && (!tlen)) {
        if ((!key) && (vlen) && (vlen<=0x9FFF)) {
            unsigned int owner=Owner->ToInt();
            if (owner<=0xFFF) {
                // max udp size
                if ((is_udp) && (vlen<=4000))
                    *is_udp=1;
                *size=vlen;
                char *ptr=new char[*size+sizeof(int)];
                *(unsigned int *)ptr=htonl(((unsigned int)*size) | (unsigned int)0xF0000000 | (owner << 16));
                char *buf=ptr+sizeof(int);
                *buffer=buf;
                memcpy(buf, Value->c_str(), vlen);
                return ptr;
            } else {
                Target->LoadBuffer("1003", 4);
                tlen=4;
            }
        } else {
            Target->LoadBuffer("1003", 4);
            tlen=4;
        }
    }

    *size=sizeof(char)+Owner->Length()+sizeof(int)+sizeof(short)+tlen+vlen;
    char *ptr=new char[*size+sizeof(int)];
    *(int *)ptr=htonl(*size);
    char *buf=ptr+sizeof(int);//new char[*size];
    *buffer=buf;

    *(unsigned char *)buf=(unsigned char)Owner->Length();
    buf+=sizeof(char);

    memcpy(buf, Owner->c_str(), Owner->Length());
    buf+=Owner->Length();

    *(int *)buf=htonl(message);
    buf+=sizeof(int);

    *(unsigned short *)buf=htons((unsigned short)tlen);
    buf+=sizeof(short);

    memcpy(buf, Target->c_str(), tlen);
    buf+=Target->Length();

    memcpy(buf, Value->c_str(), vlen);
    return ptr;
}
//-----------------------------------------------------------------------------------
unsigned int AES_encrypt(char *in_buffer, unsigned int in_buffer_size, char *out_buffer, unsigned int out_buffer_size, char *key, unsigned int key_size) {
	if (!En_inited)
		encrypt_init(key,key_size);

	return encrypt(in_buffer,in_buffer_size,out_buffer,out_buffer_size);
}
//-----------------------------------------------------------------------------------
unsigned int AES_decrypt(char *in_buffer, unsigned int in_buffer_size, char *out_buffer, unsigned int out_buffer_size, char *key, unsigned int key_size) {
	if (!Dec_inited)
		decrypt_init(key,key_size);

	return decrypt(in_buffer,in_buffer_size,out_buffer,out_buffer_size);
}
//-----------------------------------------------------------------------------------
#ifdef _WIN32
DWORD WINAPI ClientAPMWrite(LPVOID P) {
#else
void *ClientAPMWrite(void *P) {
#endif
    while (true) {
        char *block=0;
        int size=0;
#ifndef _WIN32
        if (is_writeable(pipe_out)) {
#endif
            semp(sem_apwrite);
            WriteNode *n=FIRST;

            if (n) {
                FIRST=(WriteNode *)FIRST->Next;
                block=n->buffer;
                size=n->size;
                delete n;
            }
            semv(sem_apwrite);
#ifndef _WIN32
        }
        pthread_testcancel();
#endif

        if (block) {
            if (pipe_out>0)
                write(pipe_out, block, size);
            free(block);
        } else {
#ifdef _WIN32
            Sleep(100);
#else
            usleep(100000);
#endif
        }
    }
}
//-----------------------------------------------------------------------------------
int threaded_write(int pipe_out,void *buffer, int size) {
    char *block=0;
    if (size) {
        block=(char *)malloc(size);
        if (!block)
            return -1;
        memcpy(block, buffer, size);
    } else
        return 0;

    WriteNode *n=new WriteNode;
    n->buffer=block;
    n->size=size;
    n->Next=0;

    semp(sem_apwrite);
    WriteNode *prec=FIRST;
    while (prec) {
        if (prec->Next)
            prec=(WriteNode *)prec->Next;
        else
            break;
    }
    if (prec)
        prec->Next=n;
    else
        FIRST=n;
    semv(sem_apwrite);

    return size;
}
//-----------------------------------------------------------------------------------
#ifdef _WIN32
DWORD WINAPI TimerThread(LPVOID P) {
#else
void *TimerThread(void *P) {
#endif
    while (true) {
    }
    return 0;
}
//-----------------------------------------------------------------------------------
CONCEPT_DLL_API ON_CREATE_CONTEXT MANAGEMENT_PARAMETERS {
	Invoke(INVOKE_GET_APPLICATION_INFO, HANDLER, &pipe_in, &pipe_out, &apid, &parent_apid);
	MAIN_HANDLER=HANDLER;
	seminit(sem_send, 1);
    seminit(sem_recv, 1);
    seminit(sem_apwrite, 1);
    seminit(sem_bufferlist, 1);
    send_failed=0;
    event_timeout=0;
    last_msg_seconds=0;
#ifdef _WIN32
    SECURITY_ATTRIBUTES sa = { sizeof(SECURITY_ATTRIBUTES), NULL, TRUE };
    DWORD tid;
    threadID = CreateThread(&sa, 0, ClientAPMWrite, (LPVOID)0, 0, &tid);
#else
    pthread_create(&threadID, NULL, ClientAPMWrite, (void *)0);
#endif
    memset(&remote_conceptaddr, 0, sizeof(remote_conceptaddr));
    memset(&remote_conceptudpaddr, 0, sizeof(remote_conceptudpaddr));
    return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API ON_DESTROY_CONTEXT MANAGEMENT_PARAMETERS {
    if (RTSOCKET>0) {
#ifdef _WIN32
        closesocket(RTSOCKET);
#else
        close(RTSOCKET);
#endif
        RTSOCKET=-1;
        RTCONFIRMED=0;
    }

    if (!HANDLER) {
#ifdef _WIN32
        TerminateThread(threadID, 0);
#else
        pthread_cancel(threadID);
#endif
    }
    semp(sem_apwrite);
    pipe_out=0;
    WriteNode *s_FIRST=FIRST;
    FIRST=0;
    semv(sem_apwrite);

    while (s_FIRST) {
        WriteNode *prec=s_FIRST;
        if (prec->buffer)
            free(prec->buffer);
        delete prec;

        s_FIRST=(WriteNode *)s_FIRST->Next;
    }
#ifndef _WIN32
    return (void *)"No unload";
#endif
    return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_GetAPID CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(0,"GetAPID doesn't take any parameters");
    LOCAL_INIT;

    RETURN_NUMBER(apid);
    return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_GetParentAPID CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(0,"GetParentAPID doesn't take any parameters");
    LOCAL_INIT;

    RETURN_NUMBER(parent_apid);
    return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_PeekAPMessage CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(0,"PeekAPMessage doesn't take any parameters");
    LOCAL_INIT;

    if (pipe_in<=0) {
        RETURN_NUMBER(0);
        return 0;
    }

    int res=eof(pipe_in);
    // nothing to read
    if (res<0)
        res=1;
    RETURN_NUMBER(!res);
    return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_GetAPMessage CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(3,"GetAPMessage parameters error: number GetAPMessage(var SenderAPID, var nMSG_ID, var szData)");
    LOCAL_INIT;

    CHECK_STATIC_PARAM(0,"GetAPMessage: parameter 0 should be static (number or string)");
    CHECK_STATIC_PARAM(1,"GetAPMessage: parameter 1 should be static (number or string)");
    CHECK_STATIC_PARAM(2,"GetAPMessage: parameter 2 should be static (number or string)");

    if (pipe_in<=0) {
        RETURN_NUMBER(0);
        return 0;
    }

    int params[3];
    int total=0;

    do {
        total=read(pipe_in,(void *)params, sizeof(int)*3);
        if (!total) {
#ifdef _WIN32
            Sleep(1);
#else
            usleep(1000);
#endif
        }
    } while (!total);

    if (total!=sizeof(int)*3) {
        RETURN_NUMBER(0);
        return 0;
    }

    char *data=0;
    if (params[2]>0) {
        data=new char[params[2]];
        total+=read(pipe_in,(void *)data, params[2]);
    }

    SET_NUMBER(0,params[0]);
    SET_NUMBER(1,params[1]);

    if (data) {
        SET_BUFFER(2,data,params[2]);
        delete[] data;
    } else {
        SET_STRING(2,"");
    }

    RETURN_NUMBER(total==sizeof(int)*3+params[2]);
    return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_SendAPMessage CONCEPT_API_PARAMETERS {

    PARAMETERS_CHECK(3,"SendAPMessage parameters error: number SendAPMessage(nToAppID, nMSG_ID, szData)");
    LOCAL_INIT;

    if ((pipe_out<=0) || (apid<=0)) {
        RETURN_NUMBER(0);
        return 0;
    }

    NUMBER nMSG_ID;
    NUMBER nTo;
    char *szData;
    NUMBER len;

    GET_CHECK_NUMBER(0,nTo,"SendAPMessage: parameter 0 should be a number (STATIC NUMBER)");
    GET_CHECK_NUMBER(1,nMSG_ID,"SendAPMessage: parameter 1 should be a number (STATIC NUMBER)");
    GET_CHECK_BUFFER(2,szData,len,"SendAPMessage : parameter 2 should be a string (STATIC STRING)");

    if ((int)nTo==apid) {
        TParameters *MESSAGE=new TParameters;
        MESSAGE->Sender=(long)nTo;
        MESSAGE->ID=-0x100;
        MESSAGE->Target=(long)nMSG_ID;
        if (len)
            MESSAGE->Value.LoadBuffer(szData, len);

        semp(sem_bufferlist);
        BufferedMessages.Add(MESSAGE,DATA_MESSAGE);
        semv(sem_bufferlist);

        RETURN_NUMBER(1);
        return 0;
    }

    int params[3];
    params[0]=(int)nTo;
    params[1]=(int)nMSG_ID;
    params[2]=(int)len;


    if ((params[0]<0) && (params[0]!=-1)) {
        RETURN_NUMBER(0);
        return 0;
    }

    int total=threaded_write(pipe_out,(void *)params, sizeof(int)*3);
    if (params[2]>0) {
        total+=threaded_write(pipe_out,(void *)szData, params[2]);
    }
    

    RETURN_NUMBER(total==sizeof(int)*3+params[2]);
    return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_CreateDebugPipe CONCEPT_API_PARAMETERS {
        if (PARAMETERS->COUNT!=0)
            return (void *)"CreateDebugPipe it's a reserved function. You should not use this function.";

        if (has_debug) {
            close(DebugPipes[0]);
            close(DebugPipes[1]);
            has_debug=0;
        }

        if (pipe(DebugPipes)==-1) {
            SetVariable(RESULT,VARIABLE_NUMBER,"",-1);
            return 0;
        }
        write(DebugPipes[1],(char *)&DebugPipes[1],sizeof(int));
        has_debug=0;
        SetVariable(RESULT,VARIABLE_NUMBER,"",DebugPipes[0]);
	    return 0;
}
//-----------------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_DestroyDebugPipe CONCEPT_API_PARAMETERS {
        if (PARAMETERS->COUNT!=0)
            return (void *)"DestroyDebugPipe it's a reserved function. You should not use this function.";
        if (has_debug) {
            close(DebugPipes[0]);
            close(DebugPipes[1]);
            has_debug=0;
            SetVariable(RESULT,VARIABLE_NUMBER,"",1);
        } else
            SetVariable(RESULT,VARIABLE_NUMBER,"",0);
	    return 0;
}
//-----------------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_BeginUpdateSession CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(0,"BeginUpdateSession takes no parameters");
    is_cached++;
    RETURN_NUMBER(0);
    return 0;
}
//-----------------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_EndUpdateSession CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(0,"EndUpdateSession takes no parameters");
    if (is_cached)
        is_cached--;
    if (!is_cached) {
        RETURN_NUMBER(FlushCache(CLIENT_SOCKET));
    } else {
        RETURN_NUMBER(0);
    }
    return 0;
}
//-----------------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_EndUpdateSession2 CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(0,"EndUpdateSession2 takes no parameters");
    if (is_cached)
        is_cached=0;
    if (!is_cached) {
        RETURN_NUMBER(FlushCache(CLIENT_SOCKET));
    } else {
        RETURN_NUMBER(0);
    }
    return 0;
}
//-----------------------------------------------------------------------------------
void send_message2(AnsiString Owner, AnsiString Target, int message, AnsiString message_data, char *REMOTE_PUBLIC_KEY, int CLIENT_SOCKET) {
    char *buffer=0;
    int in_content_size=0;

    SerializeBuffer(&buffer, &in_content_size, &Owner, message, &Target, &message_data, REMOTE_PUBLIC_KEY);

    if (REMOTE_PUBLIC_KEY)  {
        // criptez query-ul ...
        int out_content_size=(in_content_size/16+1)*16+5;//(((in_content_size / 48) + 1) * 64) + 256;
        char *ptr=new char[out_content_size+sizeof(int)];

        char *out_content=ptr+sizeof(int);

        semp(sem_send);
        int encrypt_result = AES_encrypt(   buffer, in_content_size,
					                        out_content, out_content_size,
                                            REMOTE_PUBLIC_KEY, 16 );
        semv(sem_send);

        if (!encrypt_result) {
            	delete[] ptr;//out_content;
                delete[] (buffer-sizeof(int));
            	return;
        }

        int size=encrypt_result;
        *(int *)ptr=htonl(size);
        deturnated_send(CLIENT_SOCKET,ptr, size+sizeof(int), 0);
        delete[] ptr;//out_content;
    } else
        deturnated_send(CLIENT_SOCKET,buffer-sizeof(int), in_content_size+sizeof(int), 0);
    delete[] (buffer-sizeof(int));
}
//-----------------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_send_message CONCEPT_API_PARAMETERS {
        // send_message( TargetName, MESSAGE_ID, MESSAGE_RESERVED, MESSAGE_DATA)
        if (PARAMETERS->COUNT!=4)
            return (void *)"send_message: function takes 4 parameters(usage: send_message [static string OWNER_NAME, static number MESSAGE_ID, string TARGET, static any MESSAGE_DATA)]";
        //if ((PARAMETERS->COUNT<4) || (PARAMETERS->COUNT>5))
        //    return (void *)"send_message: function takes 4 or 5 parameters(usage: send_message [static string OWNER_NAME, static number MESSAGE_ID, string TARGET, static any MESSAGE_DATA[, boolean threaded=false])]";
        
        char    *SENDER_NAME;
        char    *MESSAGE_DATA;
        char    *MESSAGE_TARGET;
        NUMBER  MESSAGE_ID;
        NUMBER  nDUMMY_FILL;
        char    *szDUMMY_FILL;
        INTEGER TYPE;

        AnsiString Owner;
        AnsiString Target;
        AnsiString message_data;
        //bool    threaded=false;

        // get SENDER_NAME
        GetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[0]-1],&TYPE,&SENDER_NAME,&nDUMMY_FILL);
        if (TYPE!=VARIABLE_STRING)
            return (void *)"send_message: 1st parameter should be of STATIC STRING type";
        else
            Owner.LoadBuffer(SENDER_NAME, (INTEGER)nDUMMY_FILL);

        // message ID
        GetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],&TYPE,&szDUMMY_FILL,&MESSAGE_ID);
        if (TYPE!=VARIABLE_NUMBER)
            return (void *)"send_message: 2nd parameter should be of STATIC NUMBER type";

        // get target
        GetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[2]-1],&TYPE,&MESSAGE_TARGET,&nDUMMY_FILL);
        if (TYPE!=VARIABLE_STRING)
            return (void *)"send_message: 3rd parameter should be of STATIC STRING type";
        else
            Target.LoadBuffer(MESSAGE_TARGET, (INTEGER)nDUMMY_FILL);

        // get messsage
        GetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[3]-1],&TYPE,&MESSAGE_DATA,&nDUMMY_FILL);
        if (TYPE==VARIABLE_STRING) {
            //message_data=MESSAGE_DATA;
            message_data.LoadBuffer(MESSAGE_DATA, (INTEGER)nDUMMY_FILL);
        } else
        if (TYPE==VARIABLE_NUMBER)
            message_data=AnsiString(nDUMMY_FILL);
        else
            return (void *)"send_message: 4th parameter should be of STATIC STRING or STATIC_NUMBER type";

        /*if (PARAMETERS->COUNT==5) {
            char *dummy=0;
            GetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[4]-1],&TYPE,&dummy,&nDUMMY_FILL);
            if (TYPE!=VARIABLE_NUMBER)
                return (void *)"send_message: 5th parameter should be of STATIC STRING or STATIC_NUMBER type";

            threaded=(bool)(int)nDUMMY_FILL;
        }*/

        INTEGER res=0;
        char *buffer=0;
        int in_content_size=0;
        int is_udp=0;
        int message=(int)MESSAGE_ID;

        SerializeBuffer(&buffer, &in_content_size, &Owner, message, &Target, &message_data, REMOTE_PUBLIC_KEY, &is_udp);

        if (REMOTE_PUBLIC_KEY)  {
        	// criptez query-ul ...
        	int out_content_size=(in_content_size/16+1)*16+5;//(((in_content_size / 48) + 1) * 64) + 256;
            char *ptr=new char[out_content_size+sizeof(int)];

        	char *out_content=ptr+sizeof(int);

            semp(sem_send);
        	int encrypt_result = AES_encrypt(   buffer, in_content_size,
					                            out_content, out_content_size,
                                                REMOTE_PUBLIC_KEY, 16 );
            semv(sem_send);

        	if (!encrypt_result) {
            		delete[] ptr;//out_content;
                    delete[] (buffer-sizeof(int));
            		return (void *)"send_message: <<AES_encrypt>> error encrypting message.";
        	}

            int size=encrypt_result;
            *(int *)ptr=htonl(size);

            /*if (threaded) {
                if (size>0) {
                    AnsiString *dt=new AnsiString();
                    dt->LoadBuffer(out_content, size);
                    ThreadedMessages.Add(dt, DATA_STRING);
                }
            } else*/
            if ((is_cached) && (!is_udp)) {
                Cache(out_content,size);
            } else {
                //deturnated_send(CLIENT_SOCKET,(char *)&size, sizeof(long), 0);
                //res=deturnated_send(CLIENT_SOCKET,out_content, size, 0);
                res=deturnated_send(CLIENT_SOCKET,ptr, size+sizeof(int), 0, is_udp);
            }
            delete[] ptr;//out_content;
        } else {

        	// trimit dimensiunea mai intai ...
        	//long size=SERVER_QUERY.Length()+1;
            /*if (threaded) {
            } else*/
            if (is_cached) {
                Cache(buffer,in_content_size);
            } else {
        	    /*deturnated_send(CLIENT_SOCKET,(char *)&in_content_size, sizeof(long), 0);
        	    res=deturnated_send(CLIENT_SOCKET,buffer, in_content_size, 0);*/

                res=deturnated_send(CLIENT_SOCKET,buffer-sizeof(int), in_content_size+sizeof(int), 0, is_udp);
            } 
        }
        delete[] (buffer-sizeof(int));

        SetVariable(RESULT,VARIABLE_NUMBER,"",0);
        return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_wait_message_ID CONCEPT_API_PARAMETERS {
        if ((PARAMETERS->COUNT<4) || (PARAMETERS->COUNT>5)) {
            SetVariable(RESULT,VARIABLE_NUMBER,"",0);
            return (void *)"wait_message_ID: function takes 4 to 5 parameters(usage: wait_message_ID(static string OWNER_NAME, static number MESSAGE_ID, string TARGET, static any MESSAGE_DATA[, number no_timeout=false))";
        }
        
        char do_timeout=1;
        NUMBER  nDUMMY_FILL;
        char    *szDUMMY_FILL;
        INTEGER TYPE;
        char *buffer=0;//[BUFFER_SIZE];
        INTEGER rec_count=0;
        INTEGER MSG_ID=0;
        INTEGER OWNER=0;
        AnsiString temp;

        if (send_failed) {
            int new_socket=Invoke(INVOKE_GET_PROTO_SOCKET, MAIN_HANDLER);
            if (CLIENT_SOCKET!=new_socket) {
                last_msg_seconds=time(NULL);
                CLIENT_SOCKET=new_socket;
                ResetConceptPeer(new_socket);                                                       \
                send_failed=0;
            }
        }
        if (is_cached)
            FlushCache(CLIENT_SOCKET);

        for (int i=0;i<4;i++) {
            GetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[i]-1],&TYPE,&szDUMMY_FILL,&nDUMMY_FILL);
            if ((TYPE!=VARIABLE_STRING) && (TYPE!=VARIABLE_NUMBER)) {
                SetVariable(RESULT,VARIABLE_NUMBER,"",0);
                return (void *)"wait_message_ID: all parameters are passed by reference to this function. All parameters should be of STATIC type.";
            }
            if (i==0) {
                if ((TYPE==VARIABLE_NUMBER) && (nDUMMY_FILL>0))
                    OWNER=(INTEGER)nDUMMY_FILL;
            } else
            if (i==1) {
                if (TYPE==VARIABLE_NUMBER)
                    MSG_ID=(INTEGER)nDUMMY_FILL;
                else {
                    SetVariable(RESULT,VARIABLE_NUMBER,"",0);
                    return (void *)"wait_message_ID: 2nd parameter shoult be an integer (STATIC NUMBER)";
                }
            }
        }
        if (PARAMETERS_COUNT>4) {
            GetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[4]-1],&TYPE,&szDUMMY_FILL,&nDUMMY_FILL);
            if (TYPE!=VARIABLE_NUMBER) {
                SetVariable(RESULT,VARIABLE_NUMBER,"",0);
                return (void *)"wait_message_ID: parameter 5 should be a number.";
            }
            int no_timeout=(int)nDUMMY_FILL;
            if (no_timeout)
                do_timeout=0;
        }

        SetVariable(RESULT,VARIABLE_NUMBER,"",0);

        semp(sem_bufferlist);
        int len=BufferedMessages.Count();
        if (len) {
            for (int i=0;i<len;i++) {
                TParameters *PARAM=(TParameters *)BufferedMessages.Item(i);
                if ((PARAM) && (PARAM->ID==MSG_ID) && ((OWNER<=0) || (OWNER==PARAM->Sender.ToInt()))) {
                    BufferedMessages.Remove(i);
                    semv(sem_bufferlist);

                    if (OWNER<=0) {
                        SetVariable(
                                    LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[0]-1],
                                    VARIABLE_STRING,
                                    PARAM->Sender.c_str(),
                                    PARAM->Sender.Length());//0
                    }
                    SetVariable(
                                LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],
                                VARIABLE_NUMBER,
                                "",
                                PARAM->ID);

                    SetVariable(
                                LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[2]-1],
                                VARIABLE_STRING,
                                PARAM->Target.c_str(),
                                PARAM->Target.Length());//0

                    SetVariable(
                                LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[3]-1],
                                VARIABLE_STRING,
                                PARAM->Value.c_str(),
                                PARAM->Value.Length()); //0
                    delete PARAM;

                    // setez cu true (non-0) rezultatul
                    SetVariable(RESULT,VARIABLE_NUMBER,"",i+1);
                    return 0;
                }
            }
        }
        semv(sem_bufferlist);
        // iau marimea ...
        int size=0;
        int received=0;
        unsigned int size2=0;
        char timed_out=0;
        time_t start=time(NULL);
        int cachedmessages=0;
        do {
            int seof=1;
            int count=0;
            while (seof==1) {
                seof=sock_eof(CLIENT_SOCKET);
                if (seof==1) {
#ifdef _WIN32
                    Sleep(10);
#else
                    usleep(10000);
#endif
                }
                count++;
                // 10 secounds
                if ((count>=1000) && ((MSG_ID>0) && (MSG_ID!=0x131) && (MSG_ID!=0x138))) {
                    seof=-1;
                    timed_out=1;
                    break;
                }
            }
            if (seof<=0) {
                SOCKET_RECREATE_CHECK_WAIT_MESSAGE
            }
            if ((do_timeout) && ((timed_out) && (((MSG_ID>0) && (MSG_ID!=0x131) && (MSG_ID!=0x132) && (MSG_ID!=0x137) && (MSG_ID!=0x138) && (MSG_ID!=0x139) && (MSG_ID!=0x13A)) || (cachedmessages>=50)))) {
                SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],VARIABLE_NUMBER,"",0);
                SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[2]-1],VARIABLE_STRING,"",0);
                SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[3]-1],VARIABLE_STRING,"",0);
                SetVariable(RESULT,VARIABLE_NUMBER,"",0);
                return 0;
            }

            size=0;
            size2=0;
            received=0;
            rec_count=0;
            semp(sem_recv);
            do {
                received=recv(CLIENT_SOCKET,(char *)&size+rec_count,sizeof(int)-rec_count,0);
                rec_count+=received;
            } while ((rec_count<sizeof(int)) && (received>0));

            if (!LOCAL_PRIVATE_KEY) {
		        size2=ntohl(*(unsigned int *)&size);
		        if (size2 & 0xF0000000) {
			        size = size2 & 0xFFFF;
    	        } else {
			        size = ntohl(size);
			        size2 = 0;
		        }
            } else
		        size = ntohl(size);
            //size=ntohl(size);

            if (received<=0) {
                semv(sem_recv);

                SOCKET_RECREATE_CHECK

                SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[2]-1],VARIABLE_STRING,"",0);
                SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[3]-1],VARIABLE_STRING,"",0);
                SetVariable(RESULT,VARIABLE_NUMBER,"",0);
                //return (void *)"wait_message_ID: socket disconnected or host down";
                return 0;
            }

            //if (size > BUFFER_SIZE)
            //   return (void *)"wait_message_ID: received data is too big";
            if (size) {
                buffer=new char[size+1];
                buffer[size]=0;
            }

            // iau xml-ul ...
            rec_count=0;
            do {
                int to_get=size-rec_count;
                if (to_get>CHUNK_SIZE)
                    to_get=CHUNK_SIZE;
                received=recv(CLIENT_SOCKET,buffer+rec_count,to_get,0);
                rec_count+=received;
            } while ((received>0) && (rec_count<size));

            if (received<=0) {
                semv(sem_recv);
                if (buffer)
                    delete[] buffer;

                SOCKET_RECREATE_CHECK

                SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[2]-1],VARIABLE_STRING,"",0);
                SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[3]-1],VARIABLE_STRING,"",0);
                SetVariable(RESULT,VARIABLE_NUMBER,"",0);

                //SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],VARIABLE_NUMBER,"",0x500);
                //SetVariable(RESULT,VARIABLE_NUMBER,"",0);
                //return (void *)"wait_message_ID: socket disconnected or host down";
                return 0;
            }
            semv(sem_recv);

            char *out_content=0;

            if (LOCAL_PRIVATE_KEY)  {
                out_content=new char[size*2];
                buffer[size]=0;
                int decrypt_result = AES_decrypt(   buffer, size,
                                                    out_content, size,
                                                    LOCAL_PRIVATE_KEY, /*strlen(LOCAL_PRIVATE_KEY)*/16 );
                out_content[decrypt_result]=0;
                if (!decrypt_result) {
                    delete[] out_content;
                    out_content=0;
                } else {
                    size=decrypt_result;
                    if (buffer) {
                        delete[] buffer;
                        buffer=0;
                    }
                }
            }

            int message_id;
            char *Owner;
            int OwnerLen;
            char *Target;
            int TargetLen;
            char *Value;
            int ValueLen;

            int is_looked=1;

            char buf_tmp[0xFF];
            DeSerializeBuffer2(out_content ? out_content : buffer , size, &Owner, &OwnerLen, &message_id, &Target, &TargetLen, &Value, &ValueLen, size2, buf_tmp);
            if ((message_id==MSG_ID) && (OWNER>0)) {
                if (OwnerLen) {
                    temp.LoadBuffer(Owner, OwnerLen);
                    if (OWNER!=temp.ToInt())
                        is_looked=0;
                } else
                    is_looked=0;
            }
            if ((message_id==MSG_ID) && (is_looked)) {
                if (OWNER<=0) {
                    SetVariable(
                                LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[0]-1],
                                VARIABLE_STRING,
                                Owner,
                                0);
                }
                SetVariable(
                            LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],
                            VARIABLE_NUMBER,
                            "",
                            message_id);

                SetVariable(
                            LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[2]-1],
                            VARIABLE_STRING,
                            Target,
                            TargetLen);

                SetVariable(
                            LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[3]-1],
                            VARIABLE_STRING,
                            Value,
                            ValueLen);

                if (out_content) {
                    delete[] out_content;
                    out_content=0;
                }

                if (buffer) {
                    delete[] buffer;
                    buffer=0;
                }
                SetVariable(RESULT,VARIABLE_NUMBER,"",size);
                return 0;
            } else {
                // nu este mesajul cautat, trec mai departe ...
                TParameters *MESSAGE=new TParameters;
                MESSAGE->ID=message_id;
                MESSAGE->Sender=Owner;
                if (TargetLen)
                    MESSAGE->Target.LoadBuffer(Target, TargetLen);
                if (ValueLen)
                    MESSAGE->Value.LoadBuffer(Value, ValueLen);
                semp(sem_bufferlist);
                BufferedMessages.Add(MESSAGE,DATA_MESSAGE);
                semv(sem_bufferlist);
                cachedmessages++;
            }
            if (out_content) {
                delete[] out_content;
                out_content=0;
            }

            if (buffer) {
	            delete[] buffer;
                buffer=0;
            }
            if ((do_timeout) && ((((MSG_ID>0) && (MSG_ID!=0x131) && (MSG_ID!=0x132) && (MSG_ID!=0x137) && (MSG_ID!=0x138) && (MSG_ID!=0x139) && (MSG_ID!=0x13A)) || (cachedmessages>=50)) && (time(NULL)-start>=10))) {
                SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],VARIABLE_NUMBER,"",0);
                SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[2]-1],VARIABLE_STRING,"",0);
                SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[3]-1],VARIABLE_STRING,"",0);
                SetVariable(RESULT,VARIABLE_NUMBER,"",0);
                return 0;
            }
        } while (1);
        SetVariable(RESULT,VARIABLE_NUMBER,"",size);
        return 0;
}
//-----------------------------------------------------------------------------------
int check_clientaddr(sockaddr_storage *cliaddr) {
    if (cliaddr->ss_family!=remote_conceptaddr.ss_family)
        return 0;
    if (cliaddr->ss_family == AF_INET) {
        struct sockaddr_in *s = (struct sockaddr_in *)&remote_conceptaddr;
        struct sockaddr_in *s2 = (struct sockaddr_in *)cliaddr;
        if (memcmp(&s->sin_addr, &s2->sin_addr, sizeof(s2->sin_addr)))
            return 0;
        return 1;
    } else
    if (cliaddr->ss_family == AF_INET6) {
        struct sockaddr_in6 *s = (struct sockaddr_in6 *)&remote_conceptaddr;
        struct sockaddr_in6 *s2 = (struct sockaddr_in6 *)cliaddr;
        if (memcmp(&s->sin6_addr, &s2->sin6_addr, sizeof(s2->sin6_addr)))
            return 0;
        return 1;
    }
    return 0;
}

int recv2(int socket, char *buffer, int size, int flags) {
    struct sockaddr_storage cliaddr;
    socklen_t len=sizeof(cliaddr);
    int res=recvfrom(socket, buffer, size, 0 , (struct sockaddr *)&cliaddr, &len);
    if (res>0) {
        // check host
        if (!check_clientaddr(&cliaddr))
            return 0;
        // ip changed ?
        memcpy(&remote_conceptudpaddr, &cliaddr, sizeof(cliaddr));
    }
    return res;
}

int peek(int socket) {
    struct sockaddr_storage cliaddr;
    socklen_t len=sizeof(cliaddr);
    char buffer[4008];

    int res=recvfrom(socket, buffer, sizeof(buffer), MSG_PEEK, (struct sockaddr *)&cliaddr, &len);
    if (res>0) {
        if (!check_clientaddr(&cliaddr)) {
            int res=recvfrom(socket, buffer, sizeof(buffer), 0, (struct sockaddr *)&cliaddr, &len);
            if (res<0)
                return -1;
            return 0;
        }
    }
    return res;
}

int peek_UDP_message(ParamList *PARAMETERS, VariableDATA **LOCAL_CONTEXT, VariableDATA *RESULT, CALL_BACK_VARIABLE_SET SetVariable, CALL_BACK_VARIABLE_GET GetVariable, char *LOCAL_PUBLIC_KEY, char *LOCAL_PRIVATE_KEY, char *REMOTE_PUBLIC_KEY, CALL_BACK_CLASS_MEMBER_SET CallBACKClassSet, CALL_BACK_CLASS_MEMBER_GET CallBACKClassGet, INVOKE_CALL Invoke) {
    if ((RTSOCKET<=0) || (!RTCONFIRMED)) {
        return 0;
    }
    SYSTEM_SOCKET CLIENT_SOCKET=RTSOCKET;

    if (sock_eof(CLIENT_SOCKET))
        return 0;
    int peek_check=peek(CLIENT_SOCKET);

    if (peek_check<0)
        return peek_check;
    if (!peek_check)
        return 0;

    int size=0;
    int received=0;
    int rec_count=0;
    int size2=0;
    char buf_temp[4008];
    char *buffer=0;

    semp(sem_recv);
    received=recv2(CLIENT_SOCKET, buf_temp, sizeof(buf_temp),0);
    semv(sem_recv);
    size=*(unsigned int *)buf_temp;

    if (!LOCAL_PRIVATE_KEY) {
		size2=ntohl(*(unsigned int *)&size);
		if (size2 & 0xF0000000) {
			size = size2 & 0xFFFF;
    	} else {
			size = ntohl(size);
			size2 = 0;
		}
    } else
		size = ntohl(size);

    if ((received<=0) || (size<=0) || (size>4000)/* || (size+sizeof(int)!=received)*/)
        return -1;

    if (size)
        buffer=buf_temp+sizeof(int);

    char *out_content=0;
    if (LOCAL_PRIVATE_KEY)  {
        out_content=(char *)malloc(size*2);
        buffer[size]=0;
        int decrypt_result = AES_decrypt(buffer, size, out_content, size, LOCAL_PRIVATE_KEY, 16);
        out_content[decrypt_result]=0;
        if (!decrypt_result) {
            free(out_content);
            out_content=0;
        } else
            size=decrypt_result;
    }

    AnsiString Owner;
    AnsiString Target;
    AnsiString Value;
    int message_id;

    DeSerializeBuffer(out_content ? out_content : buffer , size, &Owner, &message_id, &Target, &Value, size2);
    int valid_stream_message=0;
    if ((message_id==0x1001) && (Target==def_msg)) {
        valid_stream_message=1;
        SetVariable(
                    LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[0]-1],
                    VARIABLE_STRING,
                    Owner.c_str(),
                    Owner.Length());//0
        SetVariable(
                    LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],
                    VARIABLE_NUMBER,
                    "",
                    message_id);

        SetVariable(
                    LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[2]-1],
                    VARIABLE_STRING,
                    Target.c_str(),
                    Target.Length());//0

        SetVariable(
                    LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[3]-1],
                    VARIABLE_STRING,
                    Value.c_str(),
                    Value.Length());
    }

    if (out_content)
        free(out_content);

    // atack ?
    if (valid_stream_message) {
        SetVariable(RESULT,VARIABLE_NUMBER,"",size);
        last_msg_seconds=time(NULL);
    } else
        return -1;
    return 1;
}

CONCEPT_DLL_API CONCEPT_get_message CONCEPT_API_PARAMETERS {
        if (PARAMETERS->COUNT!=4) {
            SetVariable(RESULT,VARIABLE_NUMBER,"",0);
            return (void *)"get_message: function takes 4 parameters(usage: get_message [static string OWNER_NAME, static number MESSAGE_ID, string TARGET, static any MESSAGE_DATA)]";
        }
        if (send_failed) {
            int new_socket=Invoke(INVOKE_GET_PROTO_SOCKET, MAIN_HANDLER);
            if (CLIENT_SOCKET!=new_socket) {
                last_msg_seconds=time(NULL);
                CLIENT_SOCKET=new_socket;
                ResetConceptPeer(new_socket);                                                       \
                send_failed=0;
            } else {
                SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],VARIABLE_NUMBER,"",0x500);
                SetVariable(RESULT,VARIABLE_NUMBER,"",0);
                return 0;
            }
        }

        NUMBER  nDUMMY_FILL;
        char    *szDUMMY_FILL;
        INTEGER TYPE;
        char *buffer=0;//[BUFFER_SIZE];
        INTEGER rec_count=0;

        if (is_cached)
            FlushCache(CLIENT_SOCKET);

        // get SENDER_NAME
        for (int i=0;i<4;i++) {
            GetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[i]-1],&TYPE,&szDUMMY_FILL,&nDUMMY_FILL);
            if ((TYPE!=VARIABLE_STRING) && (TYPE!=VARIABLE_NUMBER)) {
                SetVariable(RESULT,VARIABLE_NUMBER,"",0);
                return (void *)"get_message: all parameters are passed by reference to this function. All parameters should be of STATIC type.";
            }
        }

        SetVariable(RESULT,VARIABLE_NUMBER,"",0);
        // iau marimea ...
        int size=0;
        int received=0;

        struct timeval timeout;
        timeout.tv_sec = 0;
        timeout.tv_usec = 5000;
        fd_set socks;

        int sel_val;
        int echo_sent=0;
        do {
            // verificare Buffered messages (daca am chemat wait_message_ID, este posibil sa am ceva in BuffereMessages
            semp(sem_bufferlist);
            if (BufferedMessages.Count()) {
                // setez cu true (non-0) rezultatul
                SetVariable(RESULT,VARIABLE_NUMBER,"",1);
                TParameters *PARAM=(TParameters *)BufferedMessages.Remove(0);
                semv(sem_bufferlist);
                // pregatesc setarile ...
                SetVariable(
                            LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[0]-1],
                            VARIABLE_STRING,
                            PARAM->Sender.c_str(),
                            PARAM->Sender.Length());//0
                SetVariable(
                            LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],
                            VARIABLE_NUMBER,
                            "",
                            PARAM->ID);

                SetVariable(
                            LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[2]-1],
                            VARIABLE_STRING,
                            PARAM->Target.c_str(),
                            PARAM->Target.Length());//0

                SetVariable(
                            LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[3]-1],
                            VARIABLE_STRING,
                            PARAM->Value.c_str(),
                            PARAM->Value.Length()); //0
                delete PARAM;
                last_msg_seconds=time(NULL);
                return 0;
            }
            semv(sem_bufferlist);

            FD_ZERO(&socks);
            FD_SET(CLIENT_SOCKET,&socks);

            if (pipe_in<0)
                break;
            int res_eof=eof(pipe_in);
            // nothing to read
            if (res_eof<0)
                break;//res_eof=1;

            if (!res_eof) {
                // to do ... preluare mesaj inter-application
                int params[3];
                int total=read(pipe_in,(void *)params, sizeof(int)*3);

#ifndef _WIN32
                if (total<=0) {
                   sel_val=select(FD_SETSIZE, &socks, 0, 0, &timeout);
                   if (sel_val!=0)
                      break;
                }
#endif

                if (total!=sizeof(int)*3) {
                    RETURN_NUMBER(0);
                    return 0;
                }

                char *data=0;
                if (params[2]>0) {
                    data=new char[params[2]+1];
                    total=read(pipe_in,(void *)data, params[2]);
                    data[params[2]]=0;
                }

                SetVariable(
                            LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[0]-1],
                            VARIABLE_NUMBER,
                            "",
                            params[0]);
                SetVariable(
                            LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],
                            VARIABLE_NUMBER,
                            "",
                            -0x100);
                SetVariable(
                            LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[2]-1],
                            VARIABLE_NUMBER,
                            "",
                            params[1]);

                if (data)
                    SetVariable(
                                LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[3]-1],
                                VARIABLE_STRING,
                                data,
                                params[2]);
                else
                    SetVariable(
                                LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[3]-1],
                                VARIABLE_STRING,
                                "",
                                0);

                if (data)
                    delete[] data;
                SetVariable(RESULT,VARIABLE_NUMBER,"",total);

                last_msg_seconds=time(NULL);
                return 0;
            }
            sel_val=select(FD_SETSIZE, &socks, 0, 0, &timeout);
            if (!sel_val) {
                int new_socket=Invoke(INVOKE_GET_PROTO_SOCKET, MAIN_HANDLER);
                if (CLIENT_SOCKET!=new_socket) {
                    last_msg_seconds=time(NULL);
                    CLIENT_SOCKET=new_socket;
                    ResetConceptPeer(new_socket);                                                       \
                    send_failed=0;
                }

                if ((event_timeout>0) && (last_msg_seconds)) {
                    time_t now=time(NULL);
                    if ((event_timeout>30) && (!echo_sent) && (now-last_msg_seconds > event_timeout/2)) {
                        //0x503 = client query
                        send_message2("%", "Ping", 0x503, "-1", REMOTE_PUBLIC_KEY, CLIENT_SOCKET);
                        echo_sent=1;
                    } else
                    if (now-last_msg_seconds > event_timeout) {
                        SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],VARIABLE_NUMBER,"",0x500);
                        SetVariable(RESULT,VARIABLE_NUMBER,"",0);
                        return 0;
                    }
                }
                if (send_failed) {
                    SOCKET_RECREATE_CHECK

                    SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],VARIABLE_NUMBER,"",0x500);
                    SetVariable(RESULT,VARIABLE_NUMBER,"",0);
                    return 0;
                }
// reducing CPU usage ...
#ifdef _WIN32
                Sleep(5);
#else
                usleep(5000);
#endif
            }

            if ((sel_val==0) && (RTSOCKET>0) && (RTCONFIRMED)) {
                int udp_res=peek_UDP_message(PARAMETERS, LOCAL_CONTEXT, RESULT, SetVariable, GetVariable, LOCAL_PUBLIC_KEY, LOCAL_PRIVATE_KEY, REMOTE_PUBLIC_KEY, CallBACKClassSet, CallBACKClassGet, Invoke);
                if (udp_res==1) {
                    return 0;
                } else
                if (udp_res<0) {
#ifdef _WIN32
                    closesocket(RTSOCKET);
#else
                    close(RTSOCKET);
#endif
                    RTSOCKET=-1;
                    RTCONFIRMED=0;
                }
            }
        } while (sel_val==0);

        unsigned int size2=0;
        semp(sem_recv);
        do {
            received=recv(CLIENT_SOCKET,(char *)&size+rec_count,sizeof(int)-rec_count,0);
            if (received<0) {
                semv(sem_recv);

                SOCKET_RECREATE_CHECK

                SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],VARIABLE_NUMBER,"",0x500);
                SetVariable(RESULT,VARIABLE_NUMBER,"",0);
                return 0;
            }
            rec_count+=received;
        } while ((rec_count<sizeof(int)) && (received>/*=*/0)); // received > 0
        if (!LOCAL_PRIVATE_KEY) {
		    size2=ntohl(*(unsigned int *)&size);
		    if (size2 & 0xF0000000) {
			    size = size2 & 0xFFFF;
    	    } else {
			    size = ntohl(size);
			    size2 = 0;
		    }
        } else
		    size = ntohl(size);
        //size=ntohl(size);

        if ((received<=0) || (size<=0)) {
            semv(sem_recv);

            SOCKET_RECREATE_CHECK

            SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],VARIABLE_NUMBER,"",0x500);
            SetVariable(RESULT,VARIABLE_NUMBER,"",0);
            return 0;
            //return (void *)"get_message: socket disconnected or host down";
        }

        if (size) {
            buffer=(char *)malloc(size+1);
            if (!buffer) {
                semv(sem_recv);
                SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],VARIABLE_NUMBER,"",0x500);
                SetVariable(RESULT,VARIABLE_NUMBER,"",0);
                return 0;
            }
            buffer[size]=0;
        }

        rec_count=0;
        do {
            int to_get=size-rec_count;
            if (to_get>CHUNK_SIZE)
                to_get=CHUNK_SIZE;
            received=recv(CLIENT_SOCKET,buffer+rec_count,to_get,0);
            if (received<=0) {
                semv(sem_recv);
                if (buffer)
                    free(buffer);
                SOCKET_RECREATE_CHECK

                SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],VARIABLE_NUMBER,"",0x500);
                SetVariable(RESULT,VARIABLE_NUMBER,"",0);
                return 0;
            }
            rec_count+=received;
        } while ((received>/*=*/0) && (rec_count<size));

        if (received<=0) {
            semv(sem_recv);
            if (buffer)
                free(buffer);
            SOCKET_RECREATE_CHECK

            SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],VARIABLE_NUMBER,"",0x500);
            SetVariable(RESULT,VARIABLE_NUMBER,"",0);
            return 0;//(void *)"get_message: socket disconnected or host down";
        }
        semv(sem_recv);

        char *out_content=0;

        if (LOCAL_PRIVATE_KEY)  {
            out_content=(char *)malloc(size*2);
            buffer[size]=0;
            int decrypt_result = AES_decrypt(   buffer, size,
                                                out_content, size,
                                                LOCAL_PRIVATE_KEY, /*strlen(LOCAL_PRIVATE_KEY)*/16 );
            out_content[decrypt_result]=0;
            if (!decrypt_result) {
                free(out_content);
                //return (void *)"get_message: <<RSA_decrypt>> error decrypting message.";
                out_content=0;
                //out_content=buffer;
            } else
                size=decrypt_result;
        }/* else
            out_content=buffer;*/

        AnsiString Owner;
        AnsiString Target;
        AnsiString Value;
        int message_id;

        DeSerializeBuffer(out_content ? out_content : buffer , size, &Owner, &message_id, &Target, &Value, size2);

        SetVariable(
                    LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[0]-1],
                    VARIABLE_STRING,
                    Owner.c_str(),
                    Owner.Length());//0
        SetVariable(
                    LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],
                    VARIABLE_NUMBER,
                    "",
                    message_id);

        SetVariable(
                    LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[2]-1],
                    VARIABLE_STRING,
                    Target.c_str(),
                    Target.Length());//0

        SetVariable(
                    LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[3]-1],
                    VARIABLE_STRING,
                    Value.c_str(),
                    Value.Length());

        if (out_content)
            free(out_content);

        if (buffer)
	        free(buffer);

        SetVariable(RESULT,VARIABLE_NUMBER,"",size);
        last_msg_seconds=time(NULL);
        return 0;
}
//-----------------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_confirm_message CONCEPT_API_PARAMETERS {
        if (PARAMETERS->COUNT!=1)
            return (void *)"confirm_message: function takes 1 parameter. [usage: confirm_message(VALUE)]";

        char    *STR_DATA;
        NUMBER  NUMBER_DATA;
        INTEGER TYPE;

        // get SENDER_NAME
        GetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[0]-1],&TYPE,&STR_DATA,&NUMBER_DATA);
        if (TYPE!=VARIABLE_NUMBER)
            return (void *)"confirm_message: 1st parameter should be of STATIC NUMBER type";

        deturnated_send(CLIENT_SOCKET,(char *)&NUMBER_DATA, sizeof(NUMBER), 0);

        return 0;
}
//-----------------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_raw_deturnated_send CONCEPT_API_PARAMETERS {
        // send_message( TargetName, MESSAGE_ID, MESSAGE_RESERVED, MESSAGE_DATA)
        if (PARAMETERS->COUNT!=1)
            return (void *)"raw_deturnated_send: function takes 1 parameter. [usage: raw_deturnated_send(DATA)]";
        
        char    *DATA;
        NUMBER  nDUMMY_FILL;
        INTEGER TYPE;

        // get SENDER_NAME
        GetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[0]-1],&TYPE,&DATA,&nDUMMY_FILL);
        if (TYPE!=VARIABLE_STRING)
            return (void *)"raw_deturnated_send: 1st parameter should be of STATIC STRING type";

        AnsiString SERVER_QUERY(DATA);

        int size=htonl(SERVER_QUERY.Length()+1);
        deturnated_send(CLIENT_SOCKET,(char *)&size, sizeof(int), 0);
        INTEGER res=deturnated_send(CLIENT_SOCKET,SERVER_QUERY, size, 0);
        SetVariable(RESULT,VARIABLE_NUMBER,"",res);
        return 0;
}
//-----------------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_raw_receive CONCEPT_API_PARAMETERS {
        // send_message( TargetName, MESSAGE_ID, MESSAGE_RESERVED, MESSAGE_DATA)
        if (PARAMETERS->COUNT)
            return (void *)"raw_receive: doesn't take any parameters. [usage: raw_receive()]";
        
        //char    *DATA;
        //NUMBER  nDUMMY_FILL;
        //INTEGER TYPE;
        char buffer[BUFFER_SIZE];
        int rec_count=0;

        // iau marimea ...
        int size=0;
        do {
            rec_count+=recv(CLIENT_SOCKET,(char *)&size+rec_count,sizeof(int)-rec_count,0);
        } while ((rec_count<sizeof(int)) && (rec_count>-1));
        size=ntohl(size);

        if (size >= BUFFER_SIZE)
           return (void *)"raw_receive: received data is too big";
           
        // iau xml-ul ...
        rec_count=0;
        int received=0;
        do {
            received=recv(CLIENT_SOCKET,buffer+rec_count,size-rec_count,0);
            rec_count+=received;
        } while ((received>-1) && (rec_count<size));

        buffer[rec_count]=0;

        SetVariable(RESULT,VARIABLE_STRING,buffer,0);
        return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT___DEBUG_PIPEW CONCEPT_API_PARAMETERS {
        if (PARAMETERS->COUNT)
            return (void *)"__DEBUG_PIPEW: doesn't take any parameters.";
        RETURN_NUMBER(pipe_out);
        return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT___DEBUG_PIPER CONCEPT_API_PARAMETERS {
        if (PARAMETERS->COUNT)
            return (void *)"__DEBUG_PIPEW: doesn't take any parameters.";
        RETURN_NUMBER(pipe_in);
        return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_CheckPoint CONCEPT_API_PARAMETERS {
        if (PARAMETERS->COUNT!=1)
            return (void *)"CheckPoint: function takes 1 parameter: the status of checkpoint (0 do not use check points, any other values: time to check point).";

        LOCAL_INIT;
        NUMBER nvariable;

        int status;

        GET_CHECK_NUMBER(0,nvariable,"CheckPoint: status should be a number");

        status=(int)nvariable;

        Invoke(INVOKE_CHECK_POINT, status);

        RETURN_NUMBER(0);
        return 0;
}
//---------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL(HaveMessage, 0)
    int res=0;
    semp(sem_bufferlist);
    res=(BufferedMessages.Count()!=0);
    semv(sem_bufferlist);
    if (!res) {
        semp(sem_recv);
        //RETURN_NUMBER((BufferedMessages.Count()!=0) || !sock_eof(CLIENT_SOCKET))
        res=!sock_eof(CLIENT_SOCKET);
        semv(sem_recv);
    }
    RETURN_NUMBER(res);
END_IMPL
//---------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL(Hibernate, 1)
        T_NUMBER(0)

        char c=(char)PARAM_INT(0);

        Invoke(INVOKE_HIBERNATE, MAIN_HANDLER, &c);

		RETURN_NUMBER(0)
END_IMPL
//---------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL(SetApplicationEventTimeout, 1)
        T_NUMBER(0);
        // in seconds
        event_timeout=PARAM_INT(0)/1000;
        last_msg_seconds=time(NULL);
		RETURN_NUMBER(0)
END_IMPL
//---------------------------------------------------------------------------
#ifdef _WIN32
DWORD WINAPI LocalThread(LPVOID P) {
#else
void *LocalThread(void *P) {
#endif
    TimerData *td=(TimerData *)P;
    if (td) {
        int repeat=td->repeat;
        int apid=td->APID;
        AnsiString sender=AnsiString((long)td->RID);
#ifdef _WIN32
        int ms=td->ms;
#else
        int ms=td->ms*1000;
#endif
        delete td;
        if ((repeat==0) || (repeat==1)) {
#ifdef _WIN32
            Sleep(ms);
#else
            usleep(ms);
#endif
            TParameters *MESSAGE=new TParameters;
            MESSAGE->Sender=sender;
            MESSAGE->ID=0x1001; // event fired
            MESSAGE->Target="91"; // on timer

            semp(sem_bufferlist);
            BufferedMessages.Add(MESSAGE, DATA_MESSAGE);
            semv(sem_bufferlist);
        } else
        if (repeat<0) {
            do {
#ifdef _WIN32
                Sleep(ms);
#else
                usleep(ms);
#endif
                TParameters *MESSAGE=new TParameters;
                MESSAGE->Sender=sender;
                MESSAGE->ID=0x1001; // event fired
                MESSAGE->Target="91"; // on timer

                semp(sem_bufferlist);
                // max 0x8000 messages in stack
                if (BufferedMessages.Count()<0x2000)
                    BufferedMessages.Add(MESSAGE, DATA_MESSAGE);
                semv(sem_bufferlist);
            } while (true);
        } else {
            do {
#ifdef _WIN32
                Sleep(ms);
#else
                usleep(ms);
#endif
                TParameters *MESSAGE=new TParameters;
                MESSAGE->Sender=sender;
                MESSAGE->ID=0x1001; // event fired
                MESSAGE->Target="91"; // on timer

                semp(sem_bufferlist);
                // max 0x8000 messages in stack
                if (BufferedMessages.Count()<0x8000)
                    BufferedMessages.Add(MESSAGE, DATA_MESSAGE);
                semv(sem_bufferlist);
            } while (--repeat);
        }
    }
    return 0;
}
//---------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL_MINMAX_PARAMS(LocalTimer, 2, 3)
        T_NUMBER(0)
        T_NUMBER(1)
        int repeat=0;
        if (PARAMETERS_COUNT>2) {
            T_NUMBER(2)
            repeat=PARAM_INT(2);
        }
        TimerData *td=new TimerData();
        td->RID=PARAM_INT(0);
        td->ms=PARAM_INT(1);
        td->repeat=repeat;
        td->APID=apid;
#ifdef _WIN32
        SECURITY_ATTRIBUTES sa = { sizeof(SECURITY_ATTRIBUTES), NULL, TRUE };
        DWORD tid;
        LPVOID local_threadID = CreateThread(&sa, 0, LocalThread, (LPVOID)td, 0, &tid);
#else
        pthread_t local_threadID;
        pthread_create(&local_threadID, NULL, LocalThread, (void *)td);
#endif

		RETURN_NUMBER((SYS_INT)local_threadID);
END_IMPL
//---------------------------------------------------------------------------
#ifdef _WIN32
const char *inet_ntop2(int af, const void *src, char *dst, socklen_t cnt) {
    if (af == AF_INET) {
        struct sockaddr_in in;
        memset(&in, 0, sizeof(in));
        in.sin_family = AF_INET;
        memcpy(&in.sin_addr, src, sizeof(struct in_addr));
        getnameinfo((struct sockaddr *)&in, sizeof(struct sockaddr_in), dst, cnt, NULL, 0, NI_NUMERICHOST);
        return(dst);
    } else if (af == AF_INET6)   {
        struct sockaddr_in6 in;
        memset(&in, 0, sizeof(in));
        in.sin6_family = AF_INET6;
        memcpy(&in.sin6_addr, src, sizeof(struct in_addr6));
        getnameinfo((struct sockaddr *)&in, sizeof(struct sockaddr_in6), dst, cnt, NULL, 0, NI_NUMERICHOST);
        return(dst);
    }
    return(NULL);
}

#endif

CONCEPT_FUNCTION_IMPL_MINMAX_PARAMS(GetRemoteIP, 0, 1)
    struct sockaddr_storage addr;
    char        ipstr[INET6_ADDRSTRLEN];
    int         port     = 0;
    static char *unknown = "unknown";

#ifdef _WIN32
    int peerlen = sizeof(addr);
#else
    socklen_t peerlen = sizeof(addr);
#endif
    if (getpeername(CLIENT_SOCKET, (sockaddr *)&addr, &peerlen)) {
        RETURN_STRING("unknown");
        return 0;
    }

    const char *ip = 0;
    if (addr.ss_family == AF_INET) {
        struct sockaddr_in *s = (struct sockaddr_in *)&addr;
        port = ntohs(s->sin_port);
#ifdef _WIN32
        ip   = inet_ntop2(AF_INET, &s->sin_addr, ipstr, sizeof ipstr);
#else
        ip   = inet_ntop(AF_INET, &s->sin_addr, ipstr, sizeof ipstr);
#endif
    } else
    if (addr.ss_family == AF_INET6) {
        struct sockaddr_in6 *s = (struct sockaddr_in6 *)&addr;
        port = ntohs(s->sin6_port);
#ifdef _WIN32
        ip   = inet_ntop2(AF_INET6, &s->sin6_addr, ipstr, sizeof ipstr);
#else
        ip   = inet_ntop(AF_INET6, &s->sin6_addr, ipstr, sizeof ipstr);
#endif
    }
    if (!ip)
        ip = unknown;
    if (PARAMETERS_COUNT>0) {
        SET_NUMBER(0, port);
    }
    RETURN_STRING(ip);
END_IMPL
//---------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL(CloseRTSocket, 0)
    if (RTSOCKET>0) {
#ifdef _WIN32
        closesocket(RTSOCKET);
#else
        close(RTSOCKET);
#endif
        RTSOCKET=-1;
        RTCONFIRMED=0;
    }
END_IMPL
//---------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL(CreateRTSocket, 1)
    T_NUMBER(0)
    int port = PARAM_INT(0);
    int sockfd=-1;

    if ((RTSOCKET>0) || (port<0)) {
        RETURN_NUMBER(-1);
        return 0;
    }
    RTCONFIRMED=0;
    memset(&remote_conceptudpaddr, 0, sizeof(remote_conceptudpaddr));
#ifdef _WIN32
    int peerlen = sizeof(remote_conceptaddr);
#else
    socklen_t peerlen = sizeof(remote_conceptaddr);
#endif
    if (getpeername(CLIENT_SOCKET, (sockaddr *)&remote_conceptaddr, &peerlen)) {
        RETURN_NUMBER(-1);
        return 0;
    }
    struct sockaddr_in servaddr;
    sockfd=socket(AF_INET,SOCK_DGRAM,0);
    if (sockfd<=0) {
        RETURN_NUMBER(-1);
        return 0;
    } else {
        memset(&servaddr, 0, sizeof(servaddr));
        servaddr.sin_family = AF_INET;
        servaddr.sin_addr.s_addr=htonl(INADDR_ANY);
        servaddr.sin_port=htons(port);
        int flag2=0x10;
        int flag=1;
        setsockopt(sockfd, IPPROTO_IP, IP_TOS, (char *)&flag2, sizeof(flag2));
        setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, (char *)&flag, sizeof(flag));
        if (bind(sockfd,(struct sockaddr *)&servaddr,sizeof(servaddr))) {
#ifdef _WIN32
            closesocket(sockfd);
#else
            close(sockfd);
#endif
            RETURN_NUMBER(-1);
            return 0;
        }
        RTSOCKET=sockfd;
    }
    RETURN_NUMBER(RTSOCKET);
END_IMPL
//---------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL(WaitRTSocket, 1)
    int res=0;
    char buffer[0xFF];
    struct sockaddr_storage cliaddr;
    socklen_t len=sizeof(cliaddr);

    T_NUMBER(0)
    if (RTSOCKET>0) {   
        int sock_res=sock_eof_timeout(RTSOCKET, PARAM_INT(0));
        if (sock_res==0) {
            int res_peek=peek(RTSOCKET);
            if (res_peek==1) {
                int rec_res=recvfrom(RTSOCKET, buffer, 2, 0, (struct sockaddr *)&remote_conceptudpaddr, &len);
                if (rec_res==1) {
                    remote_len=len;
                    if (sendto(RTSOCKET, buffer, 1, 0, (struct sockaddr *)&remote_conceptudpaddr, len)==1) {
                        sock_res=sock_eof_timeout(RTSOCKET, PARAM_INT(0)*3);
                        if (sock_res==0) {
                            int res_peek=peek(RTSOCKET);
                            if (res_peek==1) {
                                rec_res=recvfrom(RTSOCKET, buffer, 2, 0, (struct sockaddr *)&cliaddr, &len);
                                if (rec_res==1) {
                                    res=1;
                                    RTCONFIRMED=1;
                                }
                            }
                        }
                    }
                } else
                    res=0;
            }
        }
    }
    RETURN_NUMBER(res)
END_IMPL
//---------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL(ConceptProtocolSecured, 0)
    if (REMOTE_PUBLIC_KEY)  {
        RETURN_NUMBER(1);
    } else {
        RETURN_NUMBER(0);
    }
END_IMPL
//---------------------------------------------------------------------------
